import typing as tp

import numpy as np

from static_frame.core.assign import Assign
from static_frame.core.doc_str import doc_inject
from static_frame.core.util import EMPTY_TUPLE
from static_frame.core.util import GetItemKeyType
from static_frame.core.util import NULL_SLICE

if tp.TYPE_CHECKING:
    from static_frame.core.batch import Batch  #pylint: disable = W0611 #pragma: no cover
    from static_frame.core.bus import Bus  #pylint: disable = W0611 #pragma: no cover
    from static_frame.core.frame import Frame  #pylint: disable = W0611 #pragma: no cover
    from static_frame.core.frame import FrameAsType  #pylint: disable = W0611 #pragma: no cover
    from static_frame.core.index import Index  #pylint: disable = W0611 #pragma: no cover
    from static_frame.core.series import Series  #pylint: disable = W0611 #pragma: no cover
    from static_frame.core.type_blocks import TypeBlocks  #pylint: disable = W0611 #pragma: no cover
    from static_frame.core.index_hierarchy import IndexHierarchy  #pylint: disable = W0611 #pragma: no cover

#-------------------------------------------------------------------------------

TContainer = tp.TypeVar('TContainer',
        'Index',
        'Series',
        'Frame',
        'TypeBlocks',
        'Bus',
        'Batch',
        # 'Quilt',
        'IndexHierarchy',
        )
GetItemFunc = tp.TypeVar('GetItemFunc',
        bound=tp.Callable[[GetItemKeyType], TContainer]
        )


class Interface(tp.Generic[TContainer]):
    __slots__ = EMPTY_TUPLE
    INTERFACE: tp.Tuple[str, ...] = EMPTY_TUPLE


class InterfaceGetItem(Interface[TContainer]):

    __slots__ = ('_func',)
    INTERFACE = ('__getitem__',)

    _func: tp.Callable[[GetItemKeyType], TContainer]

    def __init__(self, func: tp.Callable[[GetItemKeyType], TContainer]) -> None:
        self._func = func #type: ignore

    def __getitem__(self, key: GetItemKeyType) -> TContainer:
        return self._func(key) #type: ignore

#-------------------------------------------------------------------------------

class InterfaceSelectDuo(Interface[TContainer]):
    '''An instance to serve as an interface to all of iloc and loc
    '''

    __slots__ = (
            '_func_iloc',
            '_func_loc',
            )
    INTERFACE = ('iloc', 'loc')

    def __init__(self, *,
            func_iloc: GetItemFunc,
            func_loc: GetItemFunc) -> None:

        self._func_iloc = func_iloc
        self._func_loc = func_loc

    @property
    def iloc(self) -> InterfaceGetItem[TContainer]:
        return InterfaceGetItem(self._func_iloc)

    @property
    def loc(self) -> InterfaceGetItem[TContainer]:
        return InterfaceGetItem(self._func_loc)

class InterfaceSelectTrio(Interface[TContainer]):
    '''An instance to serve as an interface to all of iloc, loc, and __getitem__ extractors.
    '''

    __slots__ = (
            '_func_iloc',
            '_func_loc',
            '_func_getitem',
            )
    INTERFACE = ('__getitem__', 'iloc', 'loc')

    def __init__(self, *,
            func_iloc: GetItemFunc,
            func_loc: GetItemFunc,
            func_getitem: GetItemFunc,
            ) -> None:

        self._func_iloc = func_iloc
        self._func_loc = func_loc
        self._func_getitem = func_getitem

    def __getitem__(self, key: GetItemKeyType) -> tp.Any:
        '''Label-based selection.
        '''
        return self._func_getitem(key)

    @property
    def iloc(self) -> InterfaceGetItem[TContainer]:
        '''Integer-position based selection.'''
        return InterfaceGetItem(self._func_iloc)

    @property
    def loc(self) -> InterfaceGetItem[TContainer]:
        '''Label-based selection.
        '''
        return InterfaceGetItem(self._func_loc)


class InterfaceSelectQuartet(Interface[TContainer]):
    '''An instance to serve as an interface to all of iloc, loc, and __getitem__ extractors.
    '''

    __slots__ = (
            '_func_iloc',
            '_func_loc',
            '_func_getitem',
            '_func_bloc',
            )
    INTERFACE = ('__getitem__', 'iloc', 'loc', 'bloc')

    def __init__(self, *,
            func_iloc: GetItemFunc,
            func_loc: GetItemFunc,
            func_getitem: GetItemFunc,
            func_bloc: tp.Any, # not sure what is the right type
            ) -> None:

        self._func_iloc = func_iloc
        self._func_loc = func_loc
        self._func_getitem = func_getitem
        self._func_bloc = func_bloc

    def __getitem__(self, key: GetItemKeyType) -> tp.Any:
        '''Label-based selection.
        '''
        return self._func_getitem(key)

    @property
    def bloc(self) -> InterfaceGetItem[TContainer]:
        '''Boolean based assignment.'''
        return InterfaceGetItem(self._func_bloc)

    @property
    def iloc(self) -> InterfaceGetItem[TContainer]:
        '''Integer-position based assignment.'''
        return InterfaceGetItem(self._func_iloc)

    @property
    def loc(self) -> InterfaceGetItem[TContainer]:
        '''Label-based assignment.
        '''
        return InterfaceGetItem(self._func_loc)


#-------------------------------------------------------------------------------

class InterfaceAssignTrio(InterfaceSelectTrio[TContainer]):
    '''For assignment with __getitem__, iloc, loc.
    '''

    __slots__ = (
            '_func_iloc',
            '_func_loc',
            '_func_getitem',
            'delegate'
            )

    def __init__(self, *,
            func_iloc: GetItemFunc,
            func_loc: GetItemFunc,
            func_getitem: GetItemFunc,
            delegate: tp.Type[Assign]
            ) -> None:
        InterfaceSelectTrio.__init__(self,
                func_iloc=func_iloc,
                func_loc=func_loc,
                func_getitem=func_getitem,
                )
        self.delegate = delegate #pylint: disable=E0237


class InterfaceAssignQuartet(InterfaceSelectQuartet[TContainer]):
    '''For assignment with __getitem__, iloc, loc, bloc.
    '''
    __slots__ = (
            '_func_iloc',
            '_func_loc',
            '_func_getitem',
            '_func_bloc',
            'delegate'
            )

    def __init__(self, *,
            func_iloc: GetItemFunc,
            func_loc: GetItemFunc,
            func_getitem: GetItemFunc,
            func_bloc: tp.Any, # not sure what is the right type
            delegate: tp.Type[Assign]
            ) -> None:
        InterfaceSelectQuartet.__init__(self,
                func_iloc=func_iloc,
                func_loc=func_loc,
                func_getitem=func_getitem,
                func_bloc=func_bloc,
                )
        self.delegate = delegate #pylint: disable=E0237


#-------------------------------------------------------------------------------

class InterfaceAsType(Interface[TContainer]):
    '''An instance to serve as an interface to __getitem__ extractors. Used by both :obj:`Frame` and :obj:`IndexHierarchy`.
    '''

    __slots__ = ('_func_getitem',)
    INTERFACE = ('__getitem__', '__call__')

    def __init__(self,
            func_getitem: tp.Callable[[GetItemKeyType], 'FrameAsType']
            ) -> None:
        '''
        Args:
            _func_getitem: a callable that expects a _func_getitem key and returns a FrameAsType interface; for example, Frame._extract_getitem_astype.
        '''
        self._func_getitem = func_getitem

    @doc_inject(selector='selector')
    def __getitem__(self, key: GetItemKeyType) -> 'FrameAsType':
        '''Selector of columns by label.

        Args:
            key: {key_loc}
        '''
        return self._func_getitem(key)

    def __call__(self, dtype: np.dtype) -> 'Frame':
        '''
        Apply a single ``dtype`` to all columns.
        '''
        return self._func_getitem(NULL_SLICE)(dtype)



import typing as tp


from static_frame.core.util import GetItemKeyType
from static_frame.core.util import KEY_MULTIPLE_TYPES
from static_frame.core.util import NULL_SLICE


class HLocMeta(type):

    def __getitem__(cls,
            key: GetItemKeyType
            ) -> 'HLoc':
        if not isinstance(key, tuple):
            key = (key,)
        return cls(key) #type: ignore [no-any-return]

class HLoc(metaclass=HLocMeta):
    '''A wrapper for embedding hierarchical specificiations for :obj:`static_frame.IndexHierarchy` within a single axis argument of a ``loc`` selection.

    Implemented as a container of hiearchical keys that defiines NULL slices for all lower dimensions that are not defined at construction.
    '''

    STATIC = True
    __slots__ = (
            'key',
            )

    def __init__(self, key: tp.Sequence[GetItemKeyType]):
        self.key = key

    def __iter__(self) -> tp.Iterator[GetItemKeyType]:
        return self.key.__iter__()

    def __len__(self) -> int:
        return self.key.__len__()

    def __getitem__(self, key: int) -> GetItemKeyType:
        '''
        Each key reprsents a hierarchical level; if a key is not specified, the default should be to return the null slice.
        '''
        if key >= len(self.key):
            return NULL_SLICE
        return self.key.__getitem__(key)

    def has_key_multiple(self) -> bool:
        return any(isinstance(k, KEY_MULTIPLE_TYPES) for k in self.key)




import unittest
import datetime
from  itertools import product

import numpy as np

from static_frame import Index
from static_frame import IndexGO

# from static_frame import IndexHierarchy
from static_frame import Series
from static_frame import Frame

from static_frame import IndexYear
from static_frame import IndexYearGO

from static_frame import IndexYearMonth
from static_frame import IndexYearMonthGO

from static_frame import IndexDate
from static_frame import IndexDateGO

from static_frame import IndexSecond
from static_frame import IndexSecondGO

from static_frame import IndexHour
from static_frame import IndexHourGO

from static_frame import IndexMinute
from static_frame import IndexMinuteGO

from static_frame import IndexMillisecond
from static_frame import IndexMillisecondGO

from static_frame import IndexMicrosecond
from static_frame import IndexMicrosecondGO

from static_frame import IndexNanosecond
from static_frame import IndexNanosecondGO

# from static_frame import HLoc
# from static_frame import ILoc
from static_frame.core.index import _INDEX_SLOTS
from static_frame.core.index import _INDEX_GO_SLOTS
from static_frame.core.index_datetime import _dtype_to_index_cls

from static_frame.test.test_case import TestCase
from static_frame.core.exception import LocInvalid
from static_frame.core.exception import ErrorInitIndex


class TestUnit(TestCase):


    def test_index_datetime_go_config(self) -> None:

        for base, base_go in (
                (IndexYear, IndexYearGO),
                (IndexYearMonth, IndexYearMonthGO),
                (IndexDate, IndexDateGO),
                (IndexMinute, IndexMinuteGO),
                (IndexSecond, IndexSecondGO),
                (IndexMillisecond, IndexMillisecondGO),
                (IndexNanosecond, IndexNanosecondGO),
                ):
            self.assertEqual(base._MUTABLE_CONSTRUCTOR, base_go)
            self.assertEqual(base_go._IMMUTABLE_CONSTRUCTOR, base)
            self.assertEqual(base.STATIC, True)
            self.assertEqual(base_go.STATIC, False)
            self.assertEqual(len(base.__slots__), len(_INDEX_SLOTS))
            self.assertEqual(len(base_go.__slots__), len(_INDEX_GO_SLOTS))


    #---------------------------------------------------------------------------

    def test_index_date_a(self) -> None:

        index = IndexDate.from_date_range('2018-01-01', '2018-03-01')
        self.assertEqual(index.values[0], np.datetime64('2018-01-01'))
        self.assertEqual(index.values[-1], np.datetime64('2018-03-01'))
        self.assertEqual(index.loc['2018-02-22'],
                np.datetime64('2018-02-22'))


    def test_index_date_b(self) -> None:

        with self.assertRaises(Exception):
            IndexDate([3,4,5], dtype=np.int64) #pylint: disable=E1123

        idx1 = IndexDate(['2017', '2018'])
        self.assertTrue(idx1[0].__class__ == np.datetime64)
        self.assertEqual(idx1._loc_to_iloc('2018-01-01'), 1)

        idx2 = IndexDate(['2017-01', '2018-07'])
        self.assertTrue(idx2[0].__class__ == np.datetime64)
        self.assertEqual(idx2.loc['2017-01-01'],
                np.datetime64('2017-01-01'))

    def test_index_date_c(self) -> None:
        index = IndexDate.from_date_range('2017-12-15', '2018-03-15', 2)

        self.assertEqual((index == '2017').sum(), 9)
        self.assertEqual((index == '2018-02').sum(), 14)
        self.assertEqual((index == '2018').sum(), 37)

    def test_index_date_d(self) -> None:
        index = IndexDate.from_date_range('2017-12-15', '2018-03-15', 2)
        # selct by year and year month
        self.assertAlmostEqualValues(index.loc['2017'].values, #type: ignore
                np.array(['2017-12-15', '2017-12-17', '2017-12-19', '2017-12-21',
               '2017-12-23', '2017-12-25', '2017-12-27', '2017-12-29',
               '2017-12-31'], dtype='datetime64[D]'))

        self.assertAlmostEqualValues(index.loc['2018-02'].values, #type: ignore
                np.array(['2018-02-01', '2018-02-03', '2018-02-05', '2018-02-07',
               '2018-02-09', '2018-02-11', '2018-02-13', '2018-02-15',
               '2018-02-17', '2018-02-19', '2018-02-21', '2018-02-23',
               '2018-02-25', '2018-02-27'], dtype='datetime64[D]'))

        self.assertEqual(index.loc['2018-02-19'],
                np.datetime64('2018-02-19'))


    def test_index_date_e(self) -> None:
        index = IndexDate.from_date_range('2017-12-15', '2018-03-15', 2)

        post = index + np.timedelta64(2, 'D')

        self.assertEqual(post[0], np.datetime64('2017-12-17'))


    def test_index_date_f(self) -> None:
        index = IndexDate.from_date_range('2017-12-15', '2018-01-15')

        post = index + datetime.timedelta(days=10)

        self.assertEqual(post[0], np.datetime64('2017-12-25'))
        self.assertEqual(post[-1], np.datetime64('2018-01-25'))


    def test_index_date_g(self) -> None:
        index = IndexDate.from_date_range('2017-12-15', '2018-02-15')

        post = index.loc['2018':'2018-01']  # type: ignore  # https://github.com/python/typeshed/pull/3024
        self.assertEqual(len(post), 31) #type: ignore
        self.assertEqual(post[0], np.datetime64('2018-01-01')) #type: ignore
        self.assertEqual(post[-1], np.datetime64('2018-01-31')) #type: ignore


    def test_index_date_h(self) -> None:
        index = IndexDate.from_date_range('2017-12-15', '2018-02-15')

        post = index.loc['2018':'2018-01-15']  # type: ignore  # https://github.com/python/typeshed/pull/3024
        self.assertEqual(len(post), 15) #type: ignore
        self.assertEqual(post[0], np.datetime64('2018-01-01')) #type: ignore
        self.assertEqual(post[-1], np.datetime64('2018-01-15')) #type: ignore


    def test_index_date_i(self) -> None:
        index = IndexDate.from_date_range('2017-11-15', '2018-02-15')

        post = index.loc['2017-12': '2018-01']  # type: ignore  # https://github.com/python/typeshed/pull/3024
        self.assertEqual(len(post), 62) #type: ignore
        self.assertEqual(post[0], np.datetime64('2017-12-01')) #type: ignore
        self.assertEqual(post[-1], np.datetime64('2018-01-31')) #type: ignore


    def test_index_date_j(self) -> None:
        index = IndexDate.from_date_range('2017-11-15', '2018-02-15')

        post = index.loc['2017-12': '2018']  # type: ignore  # https://github.com/python/typeshed/pull/3024
        self.assertEqual(len(post), 77) #type: ignore
        self.assertEqual(post[0], np.datetime64('2017-12-01')) #type: ignore
        self.assertEqual(post[-1], np.datetime64('2018-02-15')) #type: ignore


    def test_index_date_k(self) -> None:
        index = IndexDate.from_date_range('2017-11-15', '2018-02-15')
        post = index.loc[['2017-12-10', '2018-02-06']]
        self.assertEqual(len(post), 2) #type: ignore
        self.assertEqual(post[0], np.datetime64('2017-12-10')) #type: ignore
        self.assertEqual(post[-1], np.datetime64('2018-02-06')) #type: ignore


    def test_index_date_m(self) -> None:
        index = IndexDate.from_date_range('2017-11-15', '2018-02-15')
        # NOTE: this type of selection should possibly not be permitted
        post = index.loc[['2017', '2018']]
        self.assertEqual(len(post), 93) #type: ignore
        self.assertEqual(post[0], np.datetime64('2017-11-15')) #type: ignore
        self.assertEqual(post[-1], np.datetime64('2018-02-15')) #type: ignore

    def test_index_date_n(self) -> None:
        index = IndexDate.from_date_range('2017-11-15', '2018-02-15')
        # NOTE: this type of selection should possibly not be permitted
        post = index.loc[['2017-12', '2018-02']]
        self.assertEqual(len(post), 46) #type: ignore
        self.assertEqual(post[0], np.datetime64('2017-12-01')) #type: ignore
        self.assertEqual(post[-1], np.datetime64('2018-02-15')) #type: ignore
        self.assertEqual(
            set(post.values.astype('datetime64[M]')), #type: ignore
            {np.datetime64('2018-02'), np.datetime64('2017-12')}
            )


    def test_index_date_o(self) -> None:
        index = IndexDate.from_year_month_range('2017-12', '2018-01')
        s1 = Series(range(len(index)), index=index)
        self.assertEqual(s1['2018-01':].shape, (31,)) # type: ignore
        self.assertEqual(s1[datetime.date(2018, 1, 15):].shape, (17,)) # type: ignore
        self.assertEqual(s1['2016':].shape, (0,)) # type: ignore
        self.assertEqual(s1['2017':].shape, (62,)) # type: ignore
        self.assertEqual(s1['2019':].shape, (0,)) # type: ignore


    def test_index_date_p(self) -> None:
        index = IndexDate.from_year_month_range('2017-12', '2018-01')
        s1 = Series(range(len(index)), index=index)

        with self.assertRaises(LocInvalid):
            # out of range end date
            s1['2017-12-28':'2019-01-04'] # type: ignore #pylint: disable=W0104

        with self.assertRaises(LocInvalid):
            # out of range start date
            s1['2016-01-01':'2018-01-04'] # type: ignore #pylint: disable=W0104


    def test_index_date_q(self) -> None:
        index = IndexDate(('2017-12-30', '2017-12-31', '2018-01-05'))
        s1 = Series(range(len(index)), index=index)
        # a range beyond the observed values cannot determine a match,
        self.assertEqual(s1[:'2019'].shape, (0,)) # type: ignore
        self.assertEqual(s1['2016':].shape, (0,)) # type: ignore


    #---------------------------------------------------------------------------
    def test_index_datetime_init_a(self) -> None:

        dates = [datetime.date(*x) for x in product((2017,), (4,5,), range(1, 4))]
        s1 = Series(range(len(dates)), index=IndexDate(dates))

        with self.assertRaises(ErrorInitIndex):
            index = IndexYearMonth(s1.index)

        with self.assertRaises(ErrorInitIndex):
            index = IndexYear(s1.index) #type: ignore

        # can reuse the map if going from dt64 index to normal index
        idx2 = Index(s1.index)
        self.assertTrue(id(idx2._map) == id(s1.index._map)) #type: ignore

        idx3 = IndexDate(idx2)
        self.assertTrue(id(idx3._map) == id(s1.index._map)) #type: ignore

        with self.assertRaises(ErrorInitIndex):
            index = IndexYear(idx3) #type: ignore

        # from a date to a finer resolution has to create a new map
        idx4 = IndexMinute(idx3)
        self.assertTrue(id(idx4._map) != id(s1.index._map)) #type: ignore

        # a GO has to create a new map
        idx5 = IndexGO(s1.index)
        self.assertTrue(id(idx4._map) != id(s1.index._map)) #type: ignore

        # supplying a dtype to coerce the labels
        with self.assertRaises(ErrorInitIndex):
            idx6 = Index(s1.index, dtype='datetime64[Y]')

        with self.assertRaises(ErrorInitIndex):
            idx7 = Index(s1.index.values.astype('datetime64[Y]'))

        # final resolution from a normal index
        idx8 = IndexMinute(idx2)
        self.assertTrue(id(idx8._map) != id(idx2._map))

    def test_index_date_init_b(self) -> None:
        i1 = IndexDate(('2017-12-30', '2017-12-31', '2018-01-05'), name='foo')
        i2 = IndexDate(i1)
        self.assertEqual(i1.name, i2.name)
        self.assertEqual(i2.values.tolist(),
                [datetime.date(2017, 12, 30), datetime.date(2017, 12, 31), datetime.date(2018, 1, 5)])

    #---------------------------------------------------------------------------
    def test_index_date_from_year_month_range_a(self) -> None:
        index = IndexDate.from_year_month_range('2017-12', '2018-03')

        self.assertEqual((index == '2017').sum(), 31)
        self.assertEqual((index == '2018').sum(), 90)

        self.assertEqual(
            [str(d) for d in np.unique(index.values.astype('datetime64[M]'))],
            ['2017-12', '2018-01', '2018-02', '2018-03'])


    def test_index_date_from_year_range_a(self) -> None:
        index = IndexDate.from_year_range('2016', '2018')
        self.assertEqual(len(index), 1096)
        self.assertEqual(
                [str(d) for d in np.unique(index.values.astype('datetime64[Y]'))],
                ['2016', '2017', '2018'])

        index = IndexDate.from_year_range('2016', '2018', 2)
        self.assertEqual(len(index), 548)
        self.assertEqual(
                [str(d) for d in np.unique(index.values.astype('datetime64[Y]'))],
                ['2016', '2017', '2018'])


    def test_index_date_series_a(self) -> None:

        s = Series(range(62),
                index=IndexDate.from_year_month_range('2017-12', '2018-01'))

        self.assertEqual(s.sum(), 1891)
        self.assertEqual(s.loc[s.index == '2018-01'].sum(), 1426)
        self.assertEqual(s.loc[s.index == '2017-12'].sum(), 465)

        self.assertEqual(s['2018-01-24'], 54)

        self.assertEqual(
                s['2018-01-28':].to_pairs(),  # type: ignore  # https://github.com/python/typeshed/pull/3024
                ((np.datetime64('2018-01-28'), 58), (np.datetime64('2018-01-29'), 59), (np.datetime64('2018-01-30'), 60), (np.datetime64('2018-01-31'), 61))
                )


    def test_index_year_month_a(self) -> None:
        idx1 = IndexYearMonth(('2018-01', '2018-06'))

        self.assertEqual(idx1.values.tolist(),
            [datetime.date(2018, 1, 1), datetime.date(2018, 6, 1)])


    def test_index_year_month_b(self) -> None:
        idx1 = IndexYearMonth(('2017-12', '2018-01', '2018-02', '2018-03', '2018-04'))

        post1 = idx1.loc[np.datetime64('2018-02'):]
        self.assertEqual(
                post1.values.tolist(), #type: ignore
                [datetime.date(2018, 2, 1), datetime.date(2018, 3, 1), datetime.date(2018, 4, 1)]
                )

        # a year datetime64
        post2 = idx1.loc[np.datetime64('2018'):]
        self.assertEqual(
                post2.values.tolist(), #type: ignore
                [datetime.date(2018, 1, 1), datetime.date(2018, 2, 1), datetime.date(2018, 3, 1), datetime.date(2018, 4, 1)]
                )


    def test_index_year_month_from_date_range_a(self) -> None:
        index = IndexYearMonth.from_date_range('2017-12-15', '2018-03-15')
        self.assertEqual(len(index), 4)

        index = IndexYearMonth.from_date_range('2017-12-15', '2018-03-15', 2)
        self.assertEqual(len(index), 2)

    def test_index_year_month_from_year_month_range_a(self) -> None:

        index = IndexYearMonth.from_year_month_range(
                '2017-12-15', '2018-03-15')
        self.assertAlmostEqualValues(index.values,
                np.array(['2017-12', '2018-01', '2018-02', '2018-03'],
                dtype='datetime64[M]'))

        index = IndexYearMonth.from_year_month_range('2017-12', '2018-03')
        self.assertEqual(len(index), 4)

        self.assertEqual([str(d) for d in index.values],
                ['2017-12', '2018-01', '2018-02', '2018-03'])

        index = IndexYearMonth.from_year_month_range('2017-12', '2018-03', step=2)
        self.assertEqual([str(d) for d in index], ['2017-12', '2018-02'])


    def test_index_year_month_from_year_range_a(self) -> None:

        index = IndexYearMonth.from_year_range('2010', '2018')

        self.assertEqual(len(index), 108)
        self.assertEqual(str(index.min()), '2010-01')
        self.assertEqual(str(index.max()), '2018-12')

        index = IndexYearMonth.from_year_range('2010', '2018', 6)

        self.assertEqual(
                [str(d) for d in IndexYearMonth.from_year_range('2010', '2018', 6)],
                ['2010-01', '2010-07', '2011-01', '2011-07', '2012-01', '2012-07', '2013-01', '2013-07', '2014-01', '2014-07', '2015-01', '2015-07', '2016-01', '2016-07', '2017-01', '2017-07', '2018-01', '2018-07'])


    def test_index_year_from_date_range_a(self) -> None:

        index = IndexYear.from_date_range('2014-12-15', '2018-03-15')
        self.assertEqual(len(index), 5)

        index = IndexYear.from_date_range('2014-12-15', '2018-03-15', step=2)
        self.assertEqual([str(d) for d in index.values],
                ['2014', '2016', '2018'])


    def test_index_year_from_year_month_range_a(self) -> None:

        index = IndexYear.from_year_month_range('2014-12', '2018-03')
        self.assertEqual(len(index), 5)


    def test_index_year_from_year_range_a(self) -> None:

        index = IndexYear.from_year_range('2010', '2018')
        self.assertEqual(len(index), 9)


    def test_index_year_from_year_range_b(self) -> None:

        index = IndexYearGO.from_year_range('2010', '2018')
        self.assertEqual(len(index), 9)
        index.append('2019')
        self.assertEqual(len(index), 10)
        index.append('2020')
        self.assertTrue('2020' in index)
        self.assertTrue(len(index.__slots__), 9)

        with self.assertRaises(RuntimeError):
            index.append(np.datetime64('2009-03-01'))

        index.append(np.datetime64('2009'))

        # will strip off year from a datetime object
        index.append(datetime.date(2021, 3, 15))

        self.assertEqual(index.values.tolist(),
                [datetime.date(2010, 1, 1), datetime.date(2011, 1, 1), datetime.date(2012, 1, 1), datetime.date(2013, 1, 1), datetime.date(2014, 1, 1), datetime.date(2015, 1, 1), datetime.date(2016, 1, 1), datetime.date(2017, 1, 1), datetime.date(2018, 1, 1), datetime.date(2019, 1, 1), datetime.date(2020, 1, 1), datetime.date(2009, 1, 1), datetime.date(2021, 1, 1)]
                )

    #---------------------------------------------------------------------------
    def test_index_date_loc_to_iloc_a(self) -> None:

        index = IndexDate.from_date_range('2018-01-01', '2018-03-01')

        self.assertEqual(
                index._loc_to_iloc(np.datetime64('2018-02-11')),
                41)

        self.assertEqual(index._loc_to_iloc('2018-02-11'), 41)

        self.assertEqual(
                index._loc_to_iloc(slice('2018-02-11', '2018-02-24')),
                slice(41, 55, None))

    def test_index_date_loc_to_iloc_b(self) -> None:

        index = IndexDate.from_date_range('2018-01-01', '2018-03-01')

        # with self.assertRaises(RuntimeError):
        #     _ = index.loc_to_iloc(['2017-12-01', '2018-01-01', '2018-02-01'])

        post = index._loc_to_iloc(
                ['2017-12-01', '2018-01-01', '2018-02-01'],
                partial_selection=True)
        self.assertEqual(post, [0, 31])


    def test_index_millisecond_a(self) -> None:

        msg = '''2016-04-28 04:22:12.226
2016-04-28 16:29:21.32
2016-04-28 17:36:13.733
2016-04-30 20:21:07.848
2016-05-01 00:00:33.483
2016-05-01 03:02:03.584
2016-05-01 09:26:43.185
2016-05-01 13:45:22.576
2016-05-01 15:25:46.15'''

        idx = IndexMillisecond(msg.split('\n'))
        self.assertEqual(str(idx.dtype), 'datetime64[ms]')

        self.assertEqual(idx.loc['2016-04-30T20:21:07.848'],
                np.datetime64('2016-04-30T20:21:07.848'))

        self.assertAlmostEqualValues(
                idx.loc['2016-05-01T09:26:43.185':].values,  # type: ignore  # https://github.com/python/typeshed/pull/3024
                np.array(['2016-05-01T09:26:43.185', '2016-05-01T13:45:22.576',
       '2016-05-01T15:25:46.150'], dtype='datetime64[ms]'))

        self.assertAlmostEqualValues(idx.loc['2016-05'].values, #type: ignore
                np.array(['2016-05-01T00:00:33.483', '2016-05-01T03:02:03.584',
               '2016-05-01T09:26:43.185', '2016-05-01T13:45:22.576',
               '2016-05-01T15:25:46.150'], dtype='datetime64[ms]')
                )

        self.assertEqual(idx.loc['2016-05-01T00'].values, #type: ignore
                np.array(['2016-05-01T00:00:33.483'], dtype='datetime64[ms]'))




    def test_index_millisecond_b(self) -> None:
        # integer arguments are interpreted as milliseconds from the epoch
        idx = IndexMillisecond(range(10))
        self.assertAlmostEqualValues(idx.loc['1970-01-01T00:00:00.007':].values,  # type: ignore  # https://github.com/python/typeshed/pull/3024
                np.array(['1970-01-01T00:00:00.007', '1970-01-01T00:00:00.008',
               '1970-01-01T00:00:00.009'], dtype='datetime64[ms]'))


    def test_index_second_a(self) -> None:
        # integer arguments are interpreted as seconds from the epoch
        idx = IndexSecond(range(10))
        self.assertAlmostEqualValues(idx.loc['1970-01-01T00:00:07':].values,  # type: ignore  # https://github.com/python/typeshed/pull/3024
                np.array(['1970-01-01T00:00:07', '1970-01-01T00:00:08',
               '1970-01-01T00:00:09'], dtype='datetime64[s]')
                )


    def test_index_millisecond_series_a(self) -> None:

        msg = '''2016-04-28 04:22:12.226
2016-04-28 16:29:21.32
2016-04-28 17:36:13.733
2016-04-30 20:21:07.848
2016-05-01 00:00:33.483
2016-05-01 03:02:03.584
2016-05-01 09:26:43.185
2016-05-01 13:45:22.576
2016-05-01 15:25:46.15'''

        idx = IndexMillisecond(msg.split('\n'))
        s = Series(range(9), index=idx)

        self.assertEqual(s['2016-05-01T00:00:33.483'], 4)

        self.assertEqual(s['2016-05-01T00:00:33.483':].values.tolist(),  # type: ignore  # https://github.com/python/typeshed/pull/3024
                [4, 5, 6, 7, 8])

        self.assertEqual(s['2016-05'].to_pairs(),
                ((np.datetime64('2016-05-01T00:00:33.483'), 4), (np.datetime64('2016-05-01T03:02:03.584'), 5), (np.datetime64('2016-05-01T09:26:43.185'), 6), (np.datetime64('2016-05-01T13:45:22.576'), 7), (np.datetime64('2016-05-01T15:25:46.150'), 8)))

        self.assertEqual(s['2016-05-01T09'].to_pairs(),
                ((np.datetime64('2016-05-01T09:26:43.185'), 6),))


    def test_index_millisecond_frame_a(self) -> None:

        msg = '''2016-04-28 04:22:12.226
2016-04-28 16:29:21.32
2016-04-28 17:36:13.733
2016-04-30 20:21:07.848
2016-05-01 00:00:33.483
2016-05-01 03:02:03.584
2016-05-01 09:26:43.185
2016-05-01 13:45:22.576
2016-05-01 15:25:46.15'''

        f = Frame.from_records((x, y) for x, y in enumerate(msg.split('\n')))

        idx1 = IndexMillisecond(f[1])
        self.assertAlmostEqualValues(idx1.values,
                np.array(['2016-04-28T04:22:12.226', '2016-04-28T16:29:21.320',
               '2016-04-28T17:36:13.733', '2016-04-30T20:21:07.848',
               '2016-05-01T00:00:33.483', '2016-05-01T03:02:03.584',
               '2016-05-01T09:26:43.185', '2016-05-01T13:45:22.576',
               '2016-05-01T15:25:46.150'], dtype='datetime64[ms]'))


        idx2 = IndexSecond(f[1])

        self.assertAlmostEqualValues(idx2.values,
            np.array(['2016-04-28T04:22:12', '2016-04-28T16:29:21',
           '2016-04-28T17:36:13', '2016-04-30T20:21:07',
           '2016-05-01T00:00:33', '2016-05-01T03:02:03',
           '2016-05-01T09:26:43', '2016-05-01T13:45:22',
           '2016-05-01T15:25:46'], dtype='datetime64[s]'))


        f2 = f.set_index(1, index_constructor=IndexMillisecond)
        self.assertEqual(f2.loc['2016-05', 0].values.tolist(),
                [4, 5, 6, 7, 8])


    def test_index_minute_a(self) -> None:

        idx1 = IndexMinute(('2018-01-01T03:30', '2018-01-01T03:45', '2019-01-02T03:45'))

        self.assertEqual(idx1.loc['2019'].values.tolist(), #type: ignore
                [datetime.datetime(2019, 1, 2, 3, 45)]
                )

        self.assertEqual(idx1.loc['2018-01'].values.tolist(), #type: ignore
                [datetime.datetime(2018, 1, 1, 3, 30), datetime.datetime(2018, 1, 1, 3, 45)])


    def test_index_nanosecond_a(self) -> None:

        idx1 = IndexNanosecond(('2018-01-01T03:30', '2018-01-01T03:45', '2019-01-02T03:45'))
        self.assertTrue(len(idx1.loc['2019']), 1) #type: ignore
        self.assertTrue(len(idx1.loc['2018']), 2) #type: ignore

        # NP reduces nanoseconds to integers
        self.assertEqual(idx1.values.tolist(),
                [1514777400000000000, 1514778300000000000, 1546400700000000000])


    #---------------------------------------------------------------------------

    def test_index_datetime_binary_operator_a(self) -> None:
        index = IndexDateGO.from_date_range('2018-03-12', '2018-03-15')
        index.append('2018-03-16')

        self.assertEqual((index + 2).tolist(),
                [datetime.date(2018, 3, 14), datetime.date(2018, 3, 15), datetime.date(2018, 3, 16), datetime.date(2018, 3, 17), datetime.date(2018, 3, 18)])

        with self.assertRaises(NotImplementedError):
            _ = index @ []


    def test_index_datetime_binary_operator_b(self) -> None:
        index = IndexDateGO.from_date_range('2018-03-12', '2018-03-14')
        a1 = index + Index((1, 2, 3))
        self.assertEqual(a1.tolist(),
                [datetime.date(2018, 3, 13), datetime.date(2018, 3, 15), datetime.date(2018, 3, 17)])


    #---------------------------------------------------------------------------
    def test_index_datetime_append_a(self) -> None:
        index = IndexDateGO.from_date_range('2018-03-12', '2018-03-14')
        with self.assertRaises(KeyError):
            index.append('2018-03-12')

        index.append('2018-03-11')

        self.assertEqual(index.values.tolist(),
                [datetime.date(2018, 3, 12),
                datetime.date(2018, 3, 13),
                datetime.date(2018, 3, 14),
                datetime.date(2018, 3, 11)])


    def test_index_datetime_to_pandas_a(self) -> None:

        for cls in (IndexYear, IndexYearMonth):
            index = cls.from_year_range('2010', '2018') #type: ignore
            with self.assertRaises(NotImplementedError):
                _ = index.to_pandas()


    #---------------------------------------------------------------------------

    def test_index_datetime_nanosecond_a(self) -> None:
        index1 = IndexNanosecond(('2020-01-01', '2020-02-01'))
        index2 = IndexYearMonth(index1)
        self.assertEqual(
                index2.values.tolist(),
                [datetime.date(2020, 1, 1), datetime.date(2020, 2, 1)]
                )



    #---------------------------------------------------------------------------
    def test_index_datetime_hour_a(self) -> None:
        index1 = IndexHour(('2020-01-01', '2020-02-01'))
        self.assertEqual(index1.dtype, np.dtype('<M8[h]'))
        index2 = IndexYearMonth(index1)
        self.assertEqual(
                index2.values.tolist(),
                [datetime.date(2020, 1, 1), datetime.date(2020, 2, 1)]
                )

    def test_index_datetime_hour_b(self) -> None:
        index1 = IndexHourGO(('2020-01-01', '2020-02-01'))
        index1.append('2020-03-01')
        index2 = IndexYearMonth(index1)
        self.assertEqual(
                index2.values.tolist(),
                [datetime.date(2020, 1, 1), datetime.date(2020, 2, 1), datetime.date(2020, 3, 1)]
                )

    #---------------------------------------------------------------------------
    def test_index_datetime_microsecond_a(self) -> None:
        index1 = IndexMicrosecond(('2020-01-01', '2020-02-01'))
        self.assertEqual(index1.dtype, np.dtype('<M8[us]'))
        index2 = IndexYearMonth(index1)
        self.assertEqual(
                index2.values.tolist(),
                [datetime.date(2020, 1, 1), datetime.date(2020, 2, 1)]
                )

    def test_index_datetime_microsecond_b(self) -> None:
        index1 = IndexMicrosecondGO(('2020-01-01', '2020-02-01'))
        index1.append('2020-03-01')
        self.assertEqual(index1.dtype, np.dtype('<M8[us]'))
        index2 = IndexYearMonth(index1)
        self.assertEqual(
                index2.values.tolist(),
                [datetime.date(2020, 1, 1), datetime.date(2020, 2, 1), datetime.date(2020, 3, 1)]
                )

    #---------------------------------------------------------------------------
    def test_dtype_to_index_cls_a(self) -> None:
        t1 = _dtype_to_index_cls(True, np.dtype('datetime64[D]'))
        self.assertEqual(t1, IndexDate)

        t2 = _dtype_to_index_cls(False, np.dtype('datetime64[D]'))
        self.assertEqual(t2, IndexDateGO)

        t3 = _dtype_to_index_cls(True, np.dtype('datetime64[s]'))
        self.assertEqual(t3, IndexSecond)

        t4 = _dtype_to_index_cls(False, np.dtype('datetime64[s]'))
        self.assertEqual(t4, IndexSecondGO)

        t5 = _dtype_to_index_cls(True, np.dtype('datetime64[Y]'))
        self.assertEqual(t5, IndexYear)

        t6 = _dtype_to_index_cls(False, np.dtype('datetime64[Y]'))
        self.assertEqual(t6, IndexYearGO)

    def test_dtype_to_index_cls_b(self) -> None:
        t1 = _dtype_to_index_cls(True, np.dtype(str))
        self.assertEqual(t1, Index)

        t2 = _dtype_to_index_cls(False, np.dtype(str))
        self.assertEqual(t2, IndexGO)

        t3 = _dtype_to_index_cls(True, np.dtype(float))
        self.assertEqual(t3, Index)

        t4 = _dtype_to_index_cls(False, np.dtype(float))
        self.assertEqual(t4, IndexGO)


    #---------------------------------------------------------------------------
    def test_index_datetime_astype_a(self) -> None:

        idx1 = IndexDate(('2020-01-01', '2022-05-10'))

        self.assertEqual(
            idx1.astype('datetime64[ns]').__class__, IndexNanosecond
            )

        self.assertEqual(
            idx1.astype('datetime64[Y]').__class__, IndexYear
            )
        self.assertEqual(
            idx1.astype(str).__class__, Index
            )

    def test_index_datetime_astype_b(self) -> None:

        idx1 = IndexDateGO(('2020-01-01', '2022-05-10'))

        self.assertEqual(
            idx1.astype('datetime64[ns]').__class__, IndexNanosecondGO
            )
        self.assertEqual(
            idx1.astype('datetime64[Y]').__class__, IndexYearGO
            )
        self.assertEqual(
            idx1.astype(str).__class__, IndexGO
            )

    #---------------------------------------------------------------------------

    def test_index_datetime_from_year_month_range(self) -> None:

        date_min, date_max = (np.datetime64('2007-02'), np.datetime64('2020-04'))

        idx1 = IndexYearMonth.from_year_month_range(date_min, date_max)
        self.assertEqual(len(idx1), 159)

        idx2 = IndexYear.from_year_month_range(date_min, date_max)
        self.assertEqual(len(idx2), 14)

        idx3 = IndexDate.from_year_month_range(date_min, date_max)
        self.assertEqual(len(idx3), 4838)

    def test_index_datetime_from_date_range(self) -> None:

        date_min, date_max = (np.datetime64('2007-02'), np.datetime64('2020-04'))

        # we reject year-mos when calling from_date_range
        with self.assertRaises(RuntimeError):
            _ = IndexYearMonth.from_date_range(date_min, date_max)

    #---------------------------------------------------------------------------
    def test_index_datetime_iloc_searchsorted_a(self) -> None:
        dt64 = np.datetime64
        idx = IndexDate.from_date_range('2020-01-01', '2020-01-31')
        self.assertEqual(idx.iloc_searchsorted(dt64('2020-01-05')), 4)

        self.assertEqual(
                idx.iloc_searchsorted([dt64('2020-01-05'), dt64('2020-01-31')]).tolist(),
                [4, 30]
                )
        self.assertEqual(
                idx.iloc_searchsorted([dt64('2020-01-05'), dt64('2020-01-31')], side_left=False).tolist(),
                [5, 31]
                )

    def test_index_datetime_iloc_searchsorted_b(self) -> None:
        dt64 = np.datetime64
        idx = IndexDate.from_date_range('2020-01-01', '2020-01-31')
        self.assertEqual(idx.iloc_searchsorted('2020-01-05'), 4)
        self.assertEqual(idx.iloc_searchsorted(datetime.date(2020, 1, 5)), 4)


    def test_index_datetime_loc_searchsorted_a(self) -> None:
        dt64 = np.datetime64
        idx = IndexDate.from_date_range('2020-01-01', '2020-01-31')
        self.assertEqual(idx.loc_searchsorted(dt64('2020-01-05')).tolist(), datetime.date(2020, 1, 5))

        self.assertEqual(
                idx.loc_searchsorted([dt64('2020-01-05'), dt64('2020-01-31')],
                        side_left=False,
                        fill_value=None).tolist(),
                [datetime.date(2020, 1, 6), None]
                )

    def test_index_datetime_loc_searchsorted_b(self) -> None:
        dt64 = np.datetime64
        idx = IndexDate.from_date_range('2020-01-01', '2020-01-31')
        self.assertEqual(idx.loc_searchsorted('2020-01-05').tolist(), datetime.date(2020, 1, 5))

        self.assertEqual(
                idx.loc_searchsorted(['2020-01-05', '2020-01-31'],
                        side_left=False,
                        fill_value=None).tolist(),
                [datetime.date(2020, 1, 6), None]
                )


    #---------------------------------------------------------------------------
    def test_index_datetime_loc_to_iloc_a(self) -> None:
        dt64 = np.datetime64
        idx = IndexDate.from_date_range('2020-01-01', '2020-01-31')

        self.assertEqual(idx.loc_to_iloc(['2020-01-15', '2020-01-29']),
                [14, 28])

        self.assertEqual(idx.loc_to_iloc(idx == dt64('2020-01-13')).tolist(), #type: ignore [union-attr]
                [12])
        self.assertEqual(idx.loc_to_iloc(slice('2020-01-15', '2020-01-29')),
                slice(14, 29, None))

        self.assertEqual(idx.loc_to_iloc('2020-01-29'), 28)



if __name__ == '__main__':
    unittest.main()






import typing as tp
from itertools import zip_longest
from functools import partial
from copy import deepcopy

import numpy as np

from static_frame.core.bus import Bus
from static_frame.core.container import ContainerBase
from static_frame.core.container_util import axis_window_items
from static_frame.core.display import Display
from static_frame.core.display_config import DisplayConfig
from static_frame.core.doc_str import doc_inject
from static_frame.core.exception import AxisInvalid
from static_frame.core.exception import ErrorInitQuilt
from static_frame.core.exception import NotImplementedAxis
from static_frame.core.frame import Frame
from static_frame.core.hloc import HLoc
from static_frame.core.index_base import IndexBase
from static_frame.core.index_hierarchy import IndexHierarchy
from static_frame.core.node_iter import IterNodeAxis
from static_frame.core.node_iter import IterNodeConstructorAxis
from static_frame.core.node_iter import IterNodeType
from static_frame.core.node_iter import IterNodeWindow
from static_frame.core.node_iter import IterNodeApplyType
from static_frame.core.node_selector import InterfaceGetItem
from static_frame.core.series import Series
from static_frame.core.store import Store
from static_frame.core.store import StoreConfigMapInitializer
from static_frame.core.store_client_mixin import StoreClientMixin
from static_frame.core.store_hdf5 import StoreHDF5
from static_frame.core.store_sqlite import StoreSQLite
from static_frame.core.store_xlsx import StoreXLSX
from static_frame.core.store_zip import StoreZipCSV
from static_frame.core.store_zip import StoreZipParquet
from static_frame.core.store_zip import StoreZipPickle
from static_frame.core.store_zip import StoreZipTSV
from static_frame.core.util import AnyCallable
from static_frame.core.util import array_deepcopy
from static_frame.core.util import duplicate_filter
from static_frame.core.util import get_tuple_constructor
from static_frame.core.util import GetItemKeyType
from static_frame.core.util import GetItemKeyTypeCompound
from static_frame.core.util import INT_TYPES
from static_frame.core.util import NameType
from static_frame.core.util import NULL_SLICE
from static_frame.core.util import PathSpecifier
from static_frame.core.util import concat_resolved
from static_frame.core.style_config import StyleConfig


def get_extractor(
        deepcopy_from_bus: bool,
        is_array: bool,
        memo_active: bool,
        ) -> AnyCallable:
    '''
    Args:
        memo_active: enable usage of a common memoization dictionary accross all calls to extract from this extractor.
    '''
    if deepcopy_from_bus:
        memo: tp.Optional[tp.Dict[int, tp.Any]] = None if not memo_active else {}
        if is_array:
            return partial(array_deepcopy, memo=memo)
        return partial(deepcopy, memo=memo)
    return lambda x: x

class AxisMap:
    '''
    An AxisMap is a Series where index values point to string label in a Bus.
    '''

    @staticmethod
    def get_axis_series(
            tree: tp.Dict[tp.Hashable, IndexBase],
            ) -> Series:

        index = IndexHierarchy.from_tree(tree)
        return Series(
                index.values_at_depth(0), # store the labels as series values
                index=index,
                own_index=True,
                )

    @classmethod
    def from_bus(cls,
            bus: Bus,
            axis: int,
            deepcopy_from_bus: bool,
            ) -> tp.Tuple[Series, IndexBase]:
        '''
        Given a :obj:`Bus` and an axis, derive a :obj:`Series` with an :obj:`IndexHierarchy`; also return and validate the :obj:`Index` of the opposite axis.
        '''
        # NOTE: need to extract just axis labels, not the full Frame; need new Store/Bus loaders just for label data

        extractor = get_extractor(deepcopy_from_bus, is_array=False, memo_active=False)

        tree = {}
        opposite: tp.Optional[IndexBase] = None

        for label, f in bus.items():
            if axis == 0:
                tree[label] = extractor(f.index)
                if opposite is None:
                    opposite = extractor(f.columns)
                else:
                    if not opposite.equals(f.columns):
                        raise ErrorInitQuilt('opposite axis must have equivalent indices')
            elif axis == 1:
                tree[label] = extractor(f.columns)
                if opposite is None:
                    opposite = extractor(f.index)
                else:
                    if not opposite.equals(f.index):
                        raise ErrorInitQuilt('opposite axis must have equivalent indices')
            else:
                raise AxisInvalid(f'invalid axis {axis}')
        return cls.get_axis_series(tree), opposite # type: ignore


class Quilt(ContainerBase, StoreClientMixin):
    '''
    A :obj:`Frame`-like view of the contents of a :obj:`Bus`. With the Quilt, :obj:`Frame` contained in a :obj:`Bus` can be conceived as stacking vertically (primary axis 0) or horizontally (primary axis 1). If the labels of the primary axis are unique accross all contained :obj:`Frame, ``retain_labels`` can be set to ``False`` and underlying labels are simply concatenated; otherwise, ``retain_labels`` must be set to ``True`` and an additional depth-level is added to the primary axis labels. A :obj:`Quilt` can only be created if labels of the opposite axis of all contained :obj:`Frame` are aligned.
    '''

    __slots__ = (
            '_bus',
            '_axis',
            '_axis_map',
            '_retain_labels',
            '_axis_opposite',
            '_assign_axis',
            '_columns',
            '_index',
            '_deepcopy_from_bus',
            )

    _bus: Bus
    _axis: int
    _axis_map: tp.Optional[Series]
    _axis_opposite: tp.Optional[IndexBase]
    _columns: IndexBase
    _index: IndexBase
    _assign_axis: bool

    _NDIM: int = 2

    @classmethod
    def from_frame(cls,
            frame: Frame,
            *,
            chunksize: int,
            retain_labels: bool,
            axis: int = 0,
            name: NameType = None,
            label_extractor: tp.Optional[tp.Callable[[IndexBase], tp.Hashable]] = None,
            config: StoreConfigMapInitializer = None,
            deepcopy_from_bus: bool = False,
            ) -> 'Quilt':
        '''
        Given a :obj:`Frame`, create a :obj:`Quilt` by partitioning it along the specified ``axis`` in units of ``chunksize``, where ``axis`` 0 partitions vertically (retaining aligned columns) and 1 partions horizontally (retaining aligned index).

        Args:
            label_extractor: Function that, given the partitioned index component along the specified axis, returns a string label for that chunk.
        '''
        vector = frame._index if axis == 0 else frame._columns
        vector_len = len(vector)

        starts = range(0, vector_len, chunksize)
        if len(starts) == 1:
            ends: tp.Iterable[int] = (vector_len,)
        else:
            ends = range(starts[1], vector_len, chunksize)

        if label_extractor is None:
            label_extractor = lambda x: x.iloc[0] #type: ignore

        axis_map_components: tp.Dict[tp.Hashable, IndexBase] = {}
        opposite = None

        def values() -> tp.Iterator[Frame]:
            nonlocal opposite

            for start, end in zip_longest(starts, ends, fillvalue=vector_len):
                if axis == 0: # along rows
                    f = frame.iloc[start:end]
                    label = label_extractor(f.index) #type: ignore
                    axis_map_components[label] = f.index
                    if opposite is None:
                        opposite = f.columns
                elif axis == 1: # along columns
                    f = frame.iloc[:, start:end]
                    label = label_extractor(f.columns) #type: ignore
                    axis_map_components[label] = f.columns
                    if opposite is None:
                        opposite = f.index
                else:
                    raise AxisInvalid(f'invalid axis {axis}')
                yield f.rename(label)

        name = name if name else frame.name
        bus = Bus.from_frames(values(), config=config, name=name)

        axis_map = AxisMap.get_axis_series(axis_map_components)

        return cls(bus,
                axis=axis,
                axis_map=axis_map,
                axis_opposite=opposite,
                retain_labels=retain_labels,
                deepcopy_from_bus=deepcopy_from_bus,
                )

    #---------------------------------------------------------------------------
    # constructors by data format

    @classmethod
    def _from_store(cls,
            store: Store,
            *,
            config: StoreConfigMapInitializer = None,
            axis: int = 0,
            retain_labels: bool,
            deepcopy_from_bus: bool = False,
            max_persist: tp.Optional[int] = None,
            ) -> 'Quilt':
        bus = Bus._from_store(store=store,
                config=config,
                max_persist=max_persist, # None is default
                )
        return cls(bus,
                axis=axis,
                retain_labels=retain_labels,
                deepcopy_from_bus=deepcopy_from_bus,
                )


    @classmethod
    @doc_inject(selector='quilt_constructor')
    def from_zip_tsv(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            axis: int = 0,
            retain_labels: bool,
            deepcopy_from_bus: bool = False,
            max_persist: tp.Optional[int] = None,
            ) -> 'Quilt':
        '''
        Given a file path to zipped TSV :obj:`Quilt` store, return a :obj:`Quilt` instance.

        {args}
        '''
        store = StoreZipTSV(fp)
        return cls._from_store(store,
                config=config,
                axis=axis,
                retain_labels=retain_labels,
                deepcopy_from_bus=deepcopy_from_bus,
                max_persist=max_persist,
                )

    @classmethod
    @doc_inject(selector='quilt_constructor')
    def from_zip_csv(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            axis: int = 0,
            retain_labels: bool,
            deepcopy_from_bus: bool = False,
            max_persist: tp.Optional[int] = None,
            ) -> 'Quilt':
        '''
        Given a file path to zipped CSV :obj:`Quilt` store, return a :obj:`Quilt` instance.

        {args}
        '''
        store = StoreZipCSV(fp)
        return cls._from_store(store,
                config=config,
                axis=axis,
                retain_labels=retain_labels,
                deepcopy_from_bus=deepcopy_from_bus,
                max_persist=max_persist,
                )

    @classmethod
    @doc_inject(selector='quilt_constructor')
    def from_zip_pickle(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            axis: int = 0,
            retain_labels: bool,
            deepcopy_from_bus: bool = False,
            max_persist: tp.Optional[int] = None,
            ) -> 'Quilt':
        '''
        Given a file path to zipped pickle :obj:`Quilt` store, return a :obj:`Quilt` instance.

        {args}
        '''
        store = StoreZipPickle(fp)
        return cls._from_store(store,
                config=config,
                axis=axis,
                retain_labels=retain_labels,
                deepcopy_from_bus=deepcopy_from_bus,
                max_persist=max_persist,
                )


    @classmethod
    @doc_inject(selector='quilt_constructor')
    def from_zip_parquet(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            axis: int = 0,
            retain_labels: bool,
            deepcopy_from_bus: bool = False,
            max_persist: tp.Optional[int] = None,
            ) -> 'Quilt':
        '''
        Given a file path to zipped parquet :obj:`Quilt` store, return a :obj:`Quilt` instance.

        {args}
        '''
        store = StoreZipParquet(fp)
        return cls._from_store(store,
                config=config,
                axis=axis,
                retain_labels=retain_labels,
                deepcopy_from_bus=deepcopy_from_bus,
                max_persist=max_persist,
                )


    @classmethod
    @doc_inject(selector='quilt_constructor')
    def from_xlsx(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            axis: int = 0,
            retain_labels: bool,
            deepcopy_from_bus: bool = False,
            max_persist: tp.Optional[int] = None,
            ) -> 'Quilt':
        '''
        Given a file path to an XLSX :obj:`Quilt` store, return a :obj:`Quilt` instance.

        {args}
        '''
        # how to pass configuration for multiple sheets?
        store = StoreXLSX(fp)
        return cls._from_store(store,
                config=config,
                axis=axis,
                retain_labels=retain_labels,
                deepcopy_from_bus=deepcopy_from_bus,
                max_persist=max_persist,
                )


    @classmethod
    @doc_inject(selector='quilt_constructor')
    def from_sqlite(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            axis: int = 0,
            retain_labels: bool,
            deepcopy_from_bus: bool = False,
            max_persist: tp.Optional[int] = None,
            ) -> 'Quilt':
        '''
        Given a file path to an SQLite :obj:`Quilt` store, return a :obj:`Quilt` instance.

        {args}
        '''
        store = StoreSQLite(fp)
        return cls._from_store(store,
                config=config,
                axis=axis,
                retain_labels=retain_labels,
                deepcopy_from_bus=deepcopy_from_bus,
                max_persist=max_persist,
                )


    @classmethod
    @doc_inject(selector='quilt_constructor')
    def from_hdf5(cls,
            fp: PathSpecifier,
            *,
            config: StoreConfigMapInitializer = None,
            axis: int = 0,
            retain_labels: bool,
            deepcopy_from_bus: bool = False,
            max_persist: tp.Optional[int] = None,
            ) -> 'Quilt':
        '''
        Given a file path to a HDF5 :obj:`Quilt` store, return a :obj:`Quilt` instance.

        {args}
        '''
        store = StoreHDF5(fp)
        return cls._from_store(store,
                config=config,
                axis=axis,
                retain_labels=retain_labels,
                deepcopy_from_bus=deepcopy_from_bus,
                max_persist=max_persist,
                )

    #---------------------------------------------------------------------------

    @classmethod
    def from_items(cls,
            items: tp.Iterable[tp.Tuple[tp.Hashable, Frame]],
            *,
            axis: int = 0,
            name: NameType = None,
            retain_labels: bool,
            deepcopy_from_bus: bool = False,
            ) -> 'Quilt':
        '''
        Given an iterable of pairs of label, :obj:`Frame`, create a :obj:`Quilt`.
        '''
        bus = Bus.from_items(items, name=name)
        return cls(bus,
                axis=axis,
                retain_labels=retain_labels,
                deepcopy_from_bus=deepcopy_from_bus,
                )

    @classmethod
    def from_frames(cls,
            frames: tp.Iterable[Frame],
            *,
            axis: int = 0,
            name: NameType = None,
            retain_labels: bool,
            deepcopy_from_bus: bool = False,
            ) -> 'Quilt':
        '''Return a :obj:`Quilt` from an iterable of :obj:`Frame`; labels will be drawn from :obj:`Frame.name`.
        '''
        bus = Bus.from_frames(frames, name=name)
        return cls(bus,
                axis=axis,
                retain_labels=retain_labels,
                deepcopy_from_bus=deepcopy_from_bus,
                )

    #---------------------------------------------------------------------------
    @doc_inject(selector='quilt_init')
    def __init__(self,
            bus: Bus,
            *,
            axis: int = 0,
            retain_labels: bool,
            axis_map: tp.Optional[Series] = None,
            axis_opposite: tp.Optional[IndexBase] = None,
            deepcopy_from_bus: bool = False,
            ) -> None:
        '''
        {args}
        '''
        self._bus = bus
        self._axis = axis
        self._retain_labels = retain_labels
        self._deepcopy_from_bus = deepcopy_from_bus

        if (axis_map is None) ^ (axis_opposite is None):
            raise ErrorInitQuilt('if supplying axis_map, supply axis_opposite')

        # can creation until needed
        self._axis_map = axis_map
        self._axis_opposite = axis_opposite
        # will be set with re-axis
        # self._index = None
        # self._columns = None
        self._assign_axis = True


    #---------------------------------------------------------------------------
    # deferred loading of axis info

    def _update_axis_labels(self) -> None:
        if self._axis_map is None or self._axis_opposite is None:
            self._axis_map, self._axis_opposite = AxisMap.from_bus(
                    self._bus,
                    axis=self._axis,
                    deepcopy_from_bus=self._deepcopy_from_bus,
                    )

        if self._axis == 0:
            if not self._retain_labels:
                self._index = self._axis_map.index.level_drop(1) #type: ignore
            else: # get hierarchical
                self._index = self._axis_map.index
            self._columns = self._axis_opposite
        else:
            if not self._retain_labels:
                self._columns = self._axis_map.index.level_drop(1) #type: ignore
            else:
                self._columns = self._axis_map.index
            self._index = self._axis_opposite
        self._assign_axis = False

    #---------------------------------------------------------------------------
    # name interface

    @property #type: ignore
    @doc_inject()
    def name(self) -> NameType:
        '''{}'''
        return self._bus._series._name

    def rename(self, name: NameType) -> 'Quilt':
        '''
        Return a new Quilt with an updated name attribute.
        '''
        return self.__class__(self._bus.rename(name),
                axis=self._axis,
                retain_labels=self._retain_labels,
                deepcopy_from_bus=self._deepcopy_from_bus,
                axis_map=self._axis_map,
                axis_opposite=self._axis_opposite,
                )

    #---------------------------------------------------------------------------

    def __repr__(self) -> str:
        '''Provide a display of the :obj:`Quilt` that does not realize the entire :obj:`Frame`.
        '''
        if self.name:
            header = f'{self.__class__.__name__}: {self.name}'
        else:
            header = self.__class__.__name__
        return f'<{header} at {hex(id(self))}>'

    def display(self,
            config: tp.Optional[DisplayConfig] = None,
            *,
            style_config: tp.Optional[StyleConfig] = None,
            ) -> Display:
        '''Provide a :obj:`Frame`-style display of the :obj:`Quilt`.
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return self.to_frame().display( #type: ignore
                config,
                style_config=style_config,
                )

    #---------------------------------------------------------------------------
    # accessors

    @property #type: ignore
    @doc_inject(selector='values_2d', class_name='Quilt')
    def values(self) -> np.ndarray:
        '''
        {}
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return self.to_frame().values

    @property
    def index(self) -> IndexBase:
        '''The ``IndexBase`` instance assigned for row labels.
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return self._index

    @property
    def columns(self) -> IndexBase:
        '''The ``IndexBase`` instance assigned for column labels.
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return self._columns

    #---------------------------------------------------------------------------

    @property
    def shape(self) -> tp.Tuple[int, int]:
        '''
        Return a tuple describing the shape of the underlying NumPy array.

        Returns:
            :obj:`tp.Tuple[int]`
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return len(self._index), len(self._columns)

    @property
    def ndim(self) -> int:
        '''
        Return the number of dimensions, which for a `Frame` is always 2.

        Returns:
            :obj:`int`
        '''
        return self._NDIM

    @property
    def size(self) -> int:
        '''
        Return the size of the underlying NumPy array.

        Returns:
            :obj:`int`
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return len(self._index) * len(self._columns)

    @property
    def nbytes(self) -> int:
        '''
        Return the total bytes of the underlying NumPy arrays.

        Returns:
            :obj:`int`
        '''
        # return self._blocks.nbytes
        if self._assign_axis:
            self._update_axis_labels()
        return sum(f.nbytes for _, f in self._bus.items())

    @property
    def status(self) -> Frame:
        '''
        Return a :obj:`Frame` indicating loaded status, size, bytes, and shape of all loaded :obj:`Frame` in the contained :obj:`Quilt`.
        '''
        return self._bus.status

    #---------------------------------------------------------------------------
    # dictionary-like interface

    def keys(self) -> tp.Iterable[tp.Hashable]:
        '''Iterator of column labels.
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return self._columns

    def __iter__(self) -> tp.Iterable[tp.Hashable]:
        '''
        Iterator of column labels, same as :py:meth:`Frame.keys`.
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return self._columns.__iter__()

    def __contains__(self, value: tp.Hashable) -> bool:
        '''
        Inclusion of value in column labels.
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return self._columns.__contains__(value)

    def items(self) -> tp.Iterator[tp.Tuple[tp.Hashable, Series]]:
        '''Iterator of pairs of column label and corresponding column :obj:`Series`.
        '''
        if self._assign_axis:
            self._update_axis_labels()
        yield from self._axis_series_items(axis=0) # iterate columns

    def get(self,
            key: tp.Hashable,
            default: tp.Optional[Series] = None,
            ) -> tp.Optional[Series]:
        '''
        Return the value found at the columns key, else the default if the key is not found. This method is implemented to complete the dictionary-like interface.
        '''
        if self._assign_axis:
            self._update_axis_labels()
        if key not in self._columns:
            return default
        return self.__getitem__(key) #type: ignore

    #---------------------------------------------------------------------------
    # compatibility with StoreClientMixin

    def _items_store(self) -> tp.Iterator[tp.Tuple[tp.Hashable, Frame]]:
        '''Iterator of pairs of :obj:`Quilt` label and contained :obj:`Frame`.
        '''
        yield from self._bus.items()


    #---------------------------------------------------------------------------
    # axis iterators

    def _axis_array(self, axis: int) -> tp.Iterator[np.ndarray]:
        '''Generator of arrays across an axis

        Args:
            axis: 0 iterates over columns, 1 iterates over rows
        '''
        extractor = get_extractor(
                self._deepcopy_from_bus,
                is_array=True,
                memo_active=False,
                )

        if axis == 1: # iterate over rows
            if self._axis == 0: # bus components aligned vertically
                for _, component in self._bus.items():
                    for array in component._blocks.axis_values(axis):
                        yield extractor(array)
            else: # bus components aligned horizontally
                raise NotImplementedAxis()
        elif axis == 0: # iterate over columns
            if self._axis == 1: # bus components aligned horizontally
                for _, component in self._bus.items():
                    for array in component._blocks.axis_values(axis):
                        yield extractor(array)
            else: # bus components aligned horizontally
                raise NotImplementedAxis()
        else:
            raise AxisInvalid(f'no support for axis {axis}')

    def _axis_array_items(self, axis: int) -> tp.Iterator[tp.Tuple[tp.Hashable, np.ndarray]]:
        keys = self._index if axis == 1 else self._columns
        yield from zip(keys, self._axis_array(axis))


    def _axis_tuple(self, *,
            axis: int,
            constructor: tp.Optional[tp.Type[tp.NamedTuple]] = None,
            ) -> tp.Iterator[tp.NamedTuple]:
        '''Generator of named tuples across an axis.

        Args:
            axis: 0 iterates over columns (index axis), 1 iterates over rows (column axis)
        '''
        if constructor is None:
            if axis == 1:
                labels = self._columns.values
            elif axis == 0:
                labels = self._index.values
            else:
                raise AxisInvalid(f'no support for axis {axis}')
            # uses _make method to call with iterable
            constructor = get_tuple_constructor(labels) #type: ignore
        elif (isinstance(constructor, type) and
                issubclass(constructor, tuple) and
                hasattr(constructor, '_make')):
            constructor = constructor._make #type: ignore

        assert constructor is not None

        for axis_values in self._axis_array(axis):
            yield constructor(axis_values)

    def _axis_tuple_items(self, *,
            axis: int,
            constructor: tp.Optional[tp.Type[tp.NamedTuple]] = None,
            ) -> tp.Iterator[tp.Tuple[tp.Hashable, tp.NamedTuple]]:
        keys = self._index if axis == 1 else self._columns
        yield from zip(keys, self._axis_tuple(axis=axis, constructor=constructor))


    def _axis_series(self, axis: int) -> tp.Iterator[Series]:
        '''Generator of Series across an axis
        '''
        index = self._index if axis == 0 else self._columns
        for label, axis_values in self._axis_array_items(axis):
            yield Series(axis_values, index=index, name=label, own_index=True)

    def _axis_series_items(self, axis: int) -> tp.Iterator[tp.Tuple[tp.Hashable, np.ndarray]]:
        keys = self._index if axis == 1 else self._columns
        yield from zip(keys, self._axis_series(axis=axis))

    #---------------------------------------------------------------------------
    def _axis_window_items(self, *,
            size: int,
            axis: int = 0,
            step: int = 1,
            window_sized: bool = True,
            window_func: tp.Optional[AnyCallable] = None,
            window_valid: tp.Optional[AnyCallable] = None,
            label_shift: int = 0,
            start_shift: int = 0,
            size_increment: int = 0,
            as_array: bool = False,
            ) -> tp.Iterator[tp.Tuple[tp.Hashable, tp.Any]]:
        '''Generator of index, processed-window pairs.
        '''
        # NOTE: this will use _extract, _extract_array to get results, thus we do not need an extractor
        yield from axis_window_items(
                source=self,
                size=size,
                axis=axis,
                step=step,
                window_sized=window_sized,
                window_func=window_func,
                window_valid=window_valid,
                label_shift=label_shift,
                start_shift=start_shift,
                size_increment=size_increment,
                as_array=as_array
                )

    def _axis_window(self, *,
            size: int,
            axis: int = 0,
            step: int = 1,
            window_sized: bool = True,
            window_func: tp.Optional[AnyCallable] = None,
            window_valid: tp.Optional[AnyCallable] = None,
            label_shift: int = 0,
            start_shift: int = 0,
            size_increment: int = 0,
            as_array: bool = False,
            ) -> tp.Iterator['Frame']:
        yield from (x for _, x in self._axis_window_items(
                size=size,
                axis=axis,
                step=step,
                window_sized=window_sized,
                window_func=window_func,
                window_valid=window_valid,
                label_shift=label_shift,
                start_shift=start_shift,
                size_increment=size_increment,
                as_array=as_array
                ))

    #---------------------------------------------------------------------------
    def _extract_array(self,
            row_key: GetItemKeyType = None,
            column_key: GetItemKeyType = None,
            ) -> np.ndarray:
        '''
        Extract a consolidated array based on iloc selection.
        '''
        assert self._axis_map is not None #mypy

        extractor = get_extractor(
                self._deepcopy_from_bus,
                is_array=True,
                memo_active=False,
                )

        row_key = NULL_SLICE if row_key is None else row_key
        column_key = NULL_SLICE if column_key is None else column_key

        if row_key == NULL_SLICE and column_key == NULL_SLICE:
            if len(self._bus) == 1:
                return extractor(self._bus.iloc[0].values)

            # NOTE: do not need to call extractor when concatenate is called, as a new array is always allocated.
            arrays = [f.values for _, f in self._bus.items()]
            return concat_resolved(
                    arrays,
                    axis=self._axis,
                    )

        parts: tp.List[np.ndarray] = []
        bus_keys: tp.Iterable[tp.Hashable]

        if self._axis == 0:
            sel_key = row_key
            opposite_key = column_key
        else:
            sel_key = column_key
            opposite_key = row_key

        sel_reduces = isinstance(sel_key, INT_TYPES)
        opposite_reduces = isinstance(opposite_key, INT_TYPES)

        sel = np.full(len(self._axis_map), False)
        sel[sel_key] = True

        # get ordered unique Bus labels from AxisMap Series values; cannot use .unique as need order
        axis_map_sub: tp.Union[Series, tp.Hashable] = self._axis_map.iloc[sel_key]
        if not isinstance(axis_map_sub, Series): # we have an element integer
            bus_keys = (axis_map_sub,)
        else:
            bus_keys = duplicate_filter(axis_map_sub.values)

        for key_count, key in enumerate(bus_keys):
            sel_component = sel[self._axis_map.index._loc_to_iloc(HLoc[key])]

            if self._axis == 0:
                component = self._bus.loc[key]._extract_array(sel_component, opposite_key) #type: ignore [attr-defined]
                if sel_reduces:
                    component = component[0]
            else:
                component = self._bus.loc[key]._extract_array(opposite_key, sel_component) #type: ignore [attr-defined]
                if sel_reduces:
                    if component.ndim == 1:
                        component = component[0]
                    elif component.ndim == 2:
                        component = component[NULL_SLICE, 0]

            parts.append(component)

        if len(parts) == 1:
            return extractor(parts.pop())

        # NOTE: concatenate always allocates a new array, thus no need for extractor above
        if sel_reduces or opposite_reduces:
            # NOTE: not sure if concat_resolved is needed here
            return concat_resolved(parts)
        return concat_resolved(parts, axis=self._axis)

    def _extract(self,
            row_key: GetItemKeyType = None,
            column_key: GetItemKeyType = None,
            ) -> tp.Union[Frame, Series]:
        '''
        Extract Container based on iloc selection.
        '''
        assert self._axis_map is not None #mypy

        extractor = get_extractor(
                self._deepcopy_from_bus,
                is_array=False,
                memo_active=False,
                )

        row_key = NULL_SLICE if row_key is None else row_key
        row_key_is_array = isinstance(row_key, np.ndarray)
        column_key = NULL_SLICE if column_key is None else column_key
        column_key_is_array = isinstance(column_key, np.ndarray)

        if (not row_key_is_array and row_key == NULL_SLICE
                and not column_key_is_array and column_key == NULL_SLICE):
            if self._retain_labels and self._axis == 0:
                frames = (extractor(f.relabel_level_add(index=k))
                        for k, f in self._bus.items())
            elif self._retain_labels and self._axis == 1:
                frames = (extractor(f.relabel_level_add(columns=k))
                        for k, f in self._bus.items())
            else:
                frames = (extractor(f) for _, f in self._bus.items())

            return Frame.from_concat( #type: ignore
                    frames,
                    axis=self._axis,
                    )

        parts: tp.List[tp.Any] = []
        bus_keys: tp.Iterable[tp.Hashable]

        if self._axis == 0:
            sel_key = row_key
            opposite_key = column_key
        else:
            sel_key = column_key
            opposite_key = row_key

        sel_reduces = isinstance(sel_key, INT_TYPES)

        sel = np.full(len(self._axis_map), False)
        sel[sel_key] = True

        # get ordered unique Bus labels from AxisMap Series values; cannot use .unique as need order
        axis_map_sub: tp.Union[Series, int] = self._axis_map.iloc[sel_key]
        if not isinstance(axis_map_sub, Series): # we have an element integer
            bus_keys = (axis_map_sub,)
        else:
            bus_keys = duplicate_filter(axis_map_sub.values)

        for key_count, key in enumerate(bus_keys):
            sel_component = sel[self._axis_map.index._loc_to_iloc(HLoc[key])]

            if self._axis == 0:
                component = self._bus.loc[key].iloc[sel_component, opposite_key]
                if key_count == 0:
                    component_is_series = isinstance(component, Series)
                if self._retain_labels:
                    # component might be a Series, can call the same with first arg
                    component = component.relabel_level_add(key)
                if sel_reduces: # make Frame into a Series, Series into an element
                    component = component.iloc[0]
            else:
                component = self._bus.loc[key].iloc[opposite_key, sel_component]
                if key_count == 0:
                    component_is_series = isinstance(component, Series)
                if self._retain_labels:
                    if component_is_series:
                        component = component.relabel_level_add(key)
                    else:
                        component = component.relabel_level_add(columns=key)
                if sel_reduces: # make Frame into a Series, Series into an element
                    if component_is_series:
                        component = component.iloc[0]
                    else:
                        component = component.iloc[NULL_SLICE, 0]

            parts.append(extractor(component))

        if len(parts) == 1:
            return parts.pop() #type: ignore

        # NOTE: Series/Frame from_concate will attempt to re-use ndarrays, and thus using extractor above is appropriate
        if component_is_series:
            return Series.from_concat(parts)
        return Frame.from_concat(parts, axis=self._axis) #type: ignore

    #---------------------------------------------------------------------------

    def _extract_iloc(self, key: GetItemKeyTypeCompound) -> tp.Union[Series, Frame]:
        '''
        Give a compound key, return a new Frame. This method simply handles the variabiliyt of single or compound selectors.
        '''
        if self._assign_axis:
            self._update_axis_labels()
        if isinstance(key, tuple):
            return self._extract(*key)
        return self._extract(row_key=key)

    def _compound_loc_to_iloc(self,
            key: GetItemKeyTypeCompound) -> tp.Tuple[GetItemKeyType, GetItemKeyType]:
        '''
        Given a compound iloc key, return a tuple of row, column keys. Assumes the first argument is always a row extractor.
        '''
        if isinstance(key, tuple):
            loc_row_key, loc_column_key = key
            iloc_column_key = self._columns._loc_to_iloc(loc_column_key)
        else:
            loc_row_key = key
            iloc_column_key = None

        iloc_row_key = self._index._loc_to_iloc(loc_row_key)
        return iloc_row_key, iloc_column_key

    def _extract_loc(self, key: GetItemKeyTypeCompound) -> tp.Union[Series, Frame]:
        if self._assign_axis:
            self._update_axis_labels()
        return self._extract(*self._compound_loc_to_iloc(key))

    def _compound_loc_to_getitem_iloc(self,
            key: GetItemKeyTypeCompound) -> tp.Tuple[GetItemKeyType, GetItemKeyType]:
        '''Handle a potentially compound key in the style of __getitem__. This will raise an appropriate exception if a two argument loc-style call is attempted.
        '''
        iloc_column_key = self._columns._loc_to_iloc(key)
        return None, iloc_column_key

    @doc_inject(selector='selector')
    def __getitem__(self, key: GetItemKeyType) -> tp.Union[Frame, Series]:
        '''Selector of columns by label.

        Args:
            key: {key_loc}
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return self._extract(*self._compound_loc_to_getitem_iloc(key))

    #---------------------------------------------------------------------------
    # interfaces

    @property
    def loc(self) -> InterfaceGetItem['Frame']:
        return InterfaceGetItem(self._extract_loc) #type: ignore

    @property
    def iloc(self) -> InterfaceGetItem['Frame']:
        return InterfaceGetItem(self._extract_iloc) #type: ignore

    #---------------------------------------------------------------------------
    # iterators

    @property
    def iter_array(self) -> IterNodeAxis['Quilt']:
        '''
        Iterator of :obj:`np.array`, where arrays are drawn from columns (axis=0) or rows (axis=1)
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return IterNodeAxis(
                container=self,
                function_values=self._axis_array,
                function_items=self._axis_array_items,
                yield_type=IterNodeType.VALUES,
                apply_type=IterNodeApplyType.SERIES_VALUES,
                )

    @property
    def iter_array_items(self) -> IterNodeAxis['Quilt']:
        '''
        Iterator of pairs of label, :obj:`np.array`, where arrays are drawn from columns (axis=0) or rows (axis=1)
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return IterNodeAxis(
                container=self,
                function_values=self._axis_array,
                function_items=self._axis_array_items,
                yield_type=IterNodeType.ITEMS,
                apply_type=IterNodeApplyType.SERIES_VALUES,
                )

    @property
    def iter_tuple(self) -> IterNodeConstructorAxis['Quilt']:
        '''
        Iterator of :obj:`NamedTuple`, where tuples are drawn from columns (axis=0) or rows (axis=1). An optional ``constructor`` callable can be used to provide a :obj:`NamedTuple` class (or any other constructor called with a single iterable) to be used to create each yielded axis value.
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return IterNodeConstructorAxis(
                container=self,
                function_values=self._axis_tuple,
                function_items=self._axis_tuple_items,
                yield_type=IterNodeType.VALUES,
                apply_type=IterNodeApplyType.SERIES_VALUES,
                )

    @property
    def iter_tuple_items(self) -> IterNodeConstructorAxis['Quilt']:
        '''
        Iterator of pairs of label, :obj:`NamedTuple`, where tuples are drawn from columns (axis=0) or rows (axis=1)
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return IterNodeConstructorAxis(
                container=self,
                function_values=self._axis_tuple,
                function_items=self._axis_tuple_items,
                yield_type=IterNodeType.ITEMS,
                apply_type=IterNodeApplyType.SERIES_VALUES,
                )

    @property
    def iter_series(self) -> IterNodeAxis['Quilt']:
        '''
        Iterator of :obj:`Series`, where :obj:`Series` are drawn from columns (axis=0) or rows (axis=1)
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return IterNodeAxis(
                container=self,
                function_values=self._axis_series,
                function_items=self._axis_series_items,
                yield_type=IterNodeType.VALUES,
                apply_type=IterNodeApplyType.SERIES_VALUES,
                )

    @property
    def iter_series_items(self) -> IterNodeAxis['Quilt']:
        '''
        Iterator of pairs of label, :obj:`Series`, where :obj:`Series` are drawn from columns (axis=0) or rows (axis=1)
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return IterNodeAxis(
                container=self,
                function_values=self._axis_series,
                function_items=self._axis_series_items,
                yield_type=IterNodeType.ITEMS,
                apply_type=IterNodeApplyType.SERIES_VALUES,
                )


    #---------------------------------------------------------------------------

    @property #type: ignore
    @doc_inject(selector='window')
    def iter_window(self) -> IterNodeWindow['Quilt']:
        '''
        Iterator of windowed values, where values are given as a :obj:`Frame`.

        {args}
        '''
        if self._assign_axis:
            self._update_axis_labels()
        function_values = partial(self._axis_window, as_array=False)
        function_items = partial(self._axis_window_items, as_array=False)
        return IterNodeWindow(
                container=self,
                function_values=function_values,
                function_items=function_items,
                yield_type=IterNodeType.VALUES,
                apply_type=IterNodeApplyType.SERIES_ITEMS,
                )

    @property #type: ignore
    @doc_inject(selector='window')
    def iter_window_items(self) -> IterNodeWindow['Quilt']:
        '''
        Iterator of pairs of label, windowed values, where values are given as a :obj:`Frame`.

        {args}
        '''
        if self._assign_axis:
            self._update_axis_labels()
        function_values = partial(self._axis_window, as_array=False)
        function_items = partial(self._axis_window_items, as_array=False)
        return IterNodeWindow(
                container=self,
                function_values=function_values,
                function_items=function_items,
                yield_type=IterNodeType.ITEMS,
                apply_type=IterNodeApplyType.SERIES_ITEMS,
                )

    @property #type: ignore
    @doc_inject(selector='window')
    def iter_window_array(self) -> IterNodeWindow['Quilt']:
        '''
        Iterator of windowed values, where values are given as a :obj:`np.array`.

        {args}
        '''
        if self._assign_axis:
            self._update_axis_labels()
        function_values = partial(self._axis_window, as_array=True)
        function_items = partial(self._axis_window_items, as_array=True)
        return IterNodeWindow(
                container=self,
                function_values=function_values,
                function_items=function_items,
                yield_type=IterNodeType.VALUES,
                apply_type=IterNodeApplyType.SERIES_ITEMS,
                )

    @property #type: ignore
    @doc_inject(selector='window')
    def iter_window_array_items(self) -> IterNodeWindow['Quilt']:
        '''
        Iterator of pairs of label, windowed values, where values are given as a :obj:`np.array`.

        {args}
        '''
        if self._assign_axis:
            self._update_axis_labels()
        function_values = partial(self._axis_window, as_array=True)
        function_items = partial(self._axis_window_items, as_array=True)
        return IterNodeWindow(
                container=self,
                function_values=function_values,
                function_items=function_items,
                yield_type=IterNodeType.ITEMS,
                apply_type=IterNodeApplyType.SERIES_ITEMS,
                )

    #---------------------------------------------------------------------------
    # transformations resulting in changed dimensionality
    @doc_inject(selector='head', class_name='Quilt')
    def head(self, count: int = 5) -> 'Frame':
        '''{doc}

        Args:
            {count}
        '''
        return self.iloc[:count]

    @doc_inject(selector='tail', class_name='Quilt')
    def tail(self, count: int = 5) -> 'Frame':
        '''{doc}

        Args:
            {count}
        '''
        return self.iloc[-count:]


    #---------------------------------------------------------------------------
    def to_frame(self) -> Frame:
        '''
        Return a consolidated :obj:`Frame`.
        '''
        if self._assign_axis:
            self._update_axis_labels()
        return self._extract(NULL_SLICE, NULL_SLICE) #type: ignore



import os
import sqlite3
import typing as tp
from fractions import Fraction
from contextlib import suppress
import numpy as np


# from static_frame.core.doc_str import doc_inject
from static_frame.core.frame import Frame
from static_frame.core.store import Store
from static_frame.core.store import store_coherent_non_write
from static_frame.core.store import store_coherent_write
from static_frame.core.store import StoreConfigMap
from static_frame.core.store import StoreConfigMapInitializer
from static_frame.core.util import DTYPE_BOOL
from static_frame.core.util import DTYPE_INT_KINDS
from static_frame.core.util import DTYPE_INEXACT_KINDS
from static_frame.core.util import DTYPE_STR_KINDS
from static_frame.core.util import STORE_LABEL_DEFAULT


class StoreSQLite(Store):

    _EXT: tp.FrozenSet[str] =  frozenset(('.db', '.sqlite'))
    _BYTES_ONE = b'1'

    @staticmethod
    def _dtype_to_affinity_type(
            dtype: np.dtype,
            ) -> str:
        '''
        Return a pair of writer function, Boolean, where Boolean denotes if replacements need be applied.
        '''
        kind = dtype.kind
        if dtype == DTYPE_BOOL:
            return 'BOOLEAN' # maps to NUMERIC
        elif kind in DTYPE_STR_KINDS:
            return 'TEXT'
        elif kind in DTYPE_INT_KINDS:
            return 'INTEGER'
        elif kind in DTYPE_INEXACT_KINDS:
            return 'REAL'
        return 'NONE'

    @classmethod
    def _frame_to_table(cls,
            *,
            frame: Frame,
            label: str, # can be None
            cursor: sqlite3.Cursor,
            include_columns: bool,
            include_index: bool,
            # store_filter: tp.Optional[StoreFilter]
            ) -> None:

        # here we provide a row-based represerntation that is externally usable as an slqite db; an alternative approach would be to store one cell pre column, where the column iststored as as binary BLOB; see here https://stackoverflow.com/questions/18621513/python-insert-numpy-array-into-sqlite3-database
        field_names, dtypes = cls.get_field_names_and_dtypes(
                frame=frame,
                include_index=include_index,
                include_index_name=True,
                include_columns=include_columns,
                include_columns_name=False,
                force_brackets=True # needed for having numbers as field names
                )

        index = frame._index
        columns = frame._columns

        if not include_index:
            create_primary_key = ''
        else:
            primary_fields = ', '.join(field_names[:index.depth])
            # need leading comma
            create_primary_key = f', PRIMARY KEY ({primary_fields})'

        field_name_to_field_type = (
                (field, cls._dtype_to_affinity_type(dtype))
                for field, dtype in zip(field_names, dtypes)
                )

        create_fields = ', '.join(f'{k} {v}' for k, v in field_name_to_field_type)
        create = f'CREATE TABLE "{label}" ({create_fields}{create_primary_key})'
        cursor.execute(create)

        # works for IndexHierarchy too
        insert_fields = ', '.join(f'{k}' for k in field_names)
        insert_template = ', '.join('?' for _ in field_names)
        insert = f'INSERT INTO "{label}" ({insert_fields}) VALUES ({insert_template})'

        values = cls._get_row_iterator(frame=frame, include_index=include_index)
        cursor.executemany(insert, values())

    @store_coherent_write
    def write(self,
            items: tp.Iterable[tp.Tuple[tp.Hashable, Frame]],
            *,
            config: StoreConfigMapInitializer = None,
            # store_filter: tp.Optional[StoreFilter] = STORE_FILTER_DEFAULT,
            ) -> None:

        config_map = StoreConfigMap.from_initializer(config)

        # NOTE: register adapters for NP types:
        # numpy types go in as blobs if they are not individually converted tp python types
        sqlite3.register_adapter(np.int64, int)
        sqlite3.register_adapter(np.int32, int)
        sqlite3.register_adapter(np.int16, int)
        sqlite3.register_adapter(np.bool_, bool)
        # common python types
        sqlite3.register_adapter(Fraction, str)
        sqlite3.register_adapter(complex, lambda x: f'{x.real}:{x.imag}')

        # SQLite will naturally try to update, no replace, a DB found at an FP; this is not how all other stores work, so best to remove the file first.
        with suppress(FileNotFoundError):
            os.remove(self._fp)

        # hierarchical columns might be stored as tuples
        with sqlite3.connect(self._fp, detect_types=sqlite3.PARSE_DECLTYPES) as conn:
            cursor = conn.cursor()
            for label, frame in items:
                c = config_map[label]

                # for interface compatibility with StoreXLSX, where label can be None
                if label is STORE_LABEL_DEFAULT:
                    label = 'None'
                else:
                    label = config_map.default.label_encode(label)

                self._frame_to_table(frame=frame,
                        label=label,
                        cursor=cursor,
                        include_columns=c.include_columns,
                        include_index=c.include_index,
                        # store_filter=store_filter
                        )

            conn.commit()

    @store_coherent_non_write
    def read_many(self,
            labels: tp.Iterable[tp.Hashable],
            *,
            config: StoreConfigMapInitializer = None,
            container_type: tp.Type[Frame] = Frame,
            ) -> tp.Iterator[Frame]:

        config_map = StoreConfigMap.from_initializer(config)
        sqlite3.register_converter('BOOLEAN', lambda x: x == self._BYTES_ONE)

        with sqlite3.connect(self._fp,
                detect_types=sqlite3.PARSE_DECLTYPES
                ) as conn:

            for label in labels:
                c = config_map[label]

                if label is STORE_LABEL_DEFAULT:
                    label_encoded = 'None'
                else:
                    label_encoded = config_map.default.label_encode(label)

                query = f'SELECT * from "{label_encoded}"'

                yield tp.cast(Frame, container_type.from_sql(query=query,
                        connection=conn,
                        index_depth=c.index_depth,
                        columns_depth=c.columns_depth,
                        columns_select=c.columns_select,
                        dtypes=c.dtypes,
                        name=label,
                        consolidate_blocks=c.consolidate_blocks
                        ))

    @store_coherent_non_write
    def labels(self, *,
            config: StoreConfigMapInitializer = None,
            strip_ext: bool = True,
            ) -> tp.Iterator[tp.Hashable]:

        config_map = StoreConfigMap.from_initializer(config)

        with sqlite3.connect(self._fp) as conn:
            cursor = conn.cursor()
            cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
            for row in cursor:
                yield config_map.default.label_decode(row[0])



import typing as tp
import itertools as it
import string
import os
from urllib import request

import numpy as np
import pandas as pd
import static_frame as sf


from static_frame.performance.perf_test import PerfTest



class SampleData:

    _store: tp.Dict[str, tp.Any] = {}

    URL_CSV = 'https://data.ny.gov/api/views/xe9x-a24f/rows.csv?accessType=DOWNLOAD'
    FP_CSV = '/tmp/sf_pydata_2018.csv'
    URL_JSON = 'https://jsonplaceholder.typicode.com/photos'

    @classmethod
    def create(cls) -> None:

        if not os.path.exists(cls.FP_CSV):
            with request.urlopen(cls.URL_CSV) as response:
                with open(cls.FP_CSV, 'w') as f:
                    f.write(response.read().decode('utf-8'))

        cls._store['data_csv_fp'] = cls.FP_CSV
        cls._store['data_json_url'] = cls.URL_JSON

        labels_src = list(''.join(x) for x in it.combinations(string.ascii_lowercase, 4))
        assert len(labels_src) > 10000

        index = labels_src[:10000]
        columns = labels_src[:1000]

        data_float = np.random.rand(len(index), len(columns))

        # alt floats, Bools
        data_func = [
                lambda: np.random.rand(len(index)),
                lambda: np.random.randint(-1, 1, len(index)).astype(bool)
                ]

        cls._store['index'] = index
        cls._store['index_target'] = [idx for idx in index if 'd' in idx]
        cls._store['columns'] = columns
        cls._store['columns_target'] = [c for c in columns if 'd' in c]
        cls._store['data_float'] = data_float
        cls._store['data_func'] = data_func

        cls._store['sf.FrameFloat'] = sf.Frame(data_float, index=index, columns=columns)
        cls._store['pd.FrameFloat'] = pd.DataFrame(data_float, index=index, columns=columns)

        # mypy hates this:
        data_cols = {i: data_func[i%2]() for i in range(len(columns))}  # type: ignore
        cls._store['sf.FrameMixed'] = sf.Frame.from_dict(data_cols, index=index)
        cls._store['pd.FrameMixed'] = pd.DataFrame(data_cols, index=index)

    @classmethod
    def get(cls, key: str) -> tp.Any:
        return cls._store[key]


class FloatFrameStrLabel_100_Init(PerfTest):

    NUMBER = 100

    @classmethod
    def pd(cls) -> None:
        post = pd.DataFrame(SampleData.get('data_float'),
                index=SampleData.get('index'),
                columns=SampleData.get('columns')
                )
        assert post.shape == (10000, 1000)

    @classmethod
    def sf(cls) -> None:
        post = sf.Frame(SampleData.get('data_float'),
                index=SampleData.get('index'),
                columns=SampleData.get('columns')
                )
        assert post.shape == (10000, 1000)


class FloatFrameStrLabel_101_ApplyArrayAxis1(PerfTest):

    NUMBER = 10

    @classmethod
    def pd(cls) -> None:
        frame = SampleData.get('pd.FrameFloat')
        post = frame.apply(np.sum, axis=1, raw=True)
        assert post.shape == (10000,)


    @classmethod
    def sf(cls) -> None:
        frame = SampleData.get('sf.FrameFloat')
        post = frame.iter_array(axis=1).apply(np.sum)
        assert post.shape == (10000,)

class FloatFrameStrLabel_102_ApplyArrayAxis0(PerfTest):

    NUMBER = 10

    @classmethod
    def pd(cls) -> None:
        frame = SampleData.get('pd.FrameFloat')
        post = frame.apply(np.sum, axis=0, raw=True)
        assert post.shape == (1000,)

    @classmethod
    def sf(cls) -> None:
        frame = SampleData.get('sf.FrameFloat')
        post = frame.iter_array(axis=0).apply(np.sum)
        assert post.shape == (1000,)



class FloatFrameStrLabel_103_ApplySeriesAxis1(PerfTest):

    NUMBER = 2

    @classmethod
    def pd(cls) -> None:
        frame = SampleData.get('pd.FrameFloat')
        cols = SampleData.get('columns_target')
        post = frame.apply(lambda s: s[cols].sum(), axis=1)
        assert post.shape == (10000,)

    @classmethod
    def sf(cls) -> None:
        frame = SampleData.get('sf.FrameFloat')
        cols = SampleData.get('columns_target')
        post = frame.iter_series(axis=1).apply(lambda s: s[cols].sum())
        assert post.shape == (10000,)

class FloatFrameStrLabel_104_ApplySeriesAxis0(PerfTest):

    NUMBER = 2

    @classmethod
    def pd(cls) -> None:
        frame = SampleData.get('pd.FrameFloat')
        indices = SampleData.get('index_target')
        post = frame.apply(lambda s: s[indices].sum(), axis=0)
        assert post.shape == (1000,)

    @classmethod
    def sf(cls) -> None:
        frame = SampleData.get('sf.FrameFloat')
        indices = SampleData.get('index_target')
        post = frame.iter_series(axis=0).apply(lambda s: s[indices].sum())
        assert post.shape == (1000,)

class FloatFrameStrLabel_105_SubsetConcat(PerfTest):

    NUMBER = 2

    @classmethod
    def pd(cls) -> None:
        frame = SampleData.get('pd.FrameFloat')
        post = pd.concat((
                frame[[c for c in frame.columns if c.startswith('ad')]],
                frame[[c for c in frame.columns if c.startswith('ae')]]
                ), axis=1)
        assert post.shape == (10000, 441)

    @classmethod
    def sf(cls) -> None:
        frame = SampleData.get('sf.FrameFloat')
        post = sf.Frame.from_concat((
                frame[[c for c in frame.columns if c.startswith('ad')]],
                frame[[c for c in frame.columns if c.startswith('ae')]]
                ), axis=1)
        assert post.shape == (10000, 441)







class MixedFrameIntLabel_200_Init(PerfTest):

    NUMBER = 10

    @classmethod
    def pd(cls) -> None:

        data_func = SampleData.get('data_func')
        columns = SampleData.get('columns')
        index = SampleData.get('index')
        post = pd.DataFrame({i: data_func[i%2]() for i in range(len(columns))}, index=index)
        assert post.shape == (10000, 1000)

    @classmethod
    def sf(cls) -> None:
        data_func = SampleData.get('data_func')
        columns = SampleData.get('columns')
        index = SampleData.get('index')
        post = sf.Frame.from_dict({i: data_func[i%2]() for i in range(len(columns))}, index=index)
        assert post.shape == (10000, 1000)



class MixedFrameIntLabel_201_SubsetSumAxis0(PerfTest):

    NUMBER = 10

    @classmethod
    def pd(cls) -> None:

        frame = SampleData.get('pd.FrameMixed')
        post = frame[[c for c in frame.columns if c%2 == 0]].sum()
        assert post.shape == (500,)

    @classmethod
    def sf(cls) -> None:
        frame = SampleData.get('sf.FrameMixed')
        post = frame[[c for c in frame.columns if c%2 == 0]].sum()
        assert post.shape == (500,)



class MixedFrameIntLabel_202_SubsetSumAxis1(PerfTest):

    NUMBER = 10

    @classmethod
    def pd(cls) -> None:

        frame = SampleData.get('pd.FrameMixed')
        post = frame[[c for c in frame.columns if c%2 == 0]].sum(axis=1)
        assert post.shape == (10000,)

    @classmethod
    def sf(cls) -> None:
        frame = SampleData.get('sf.FrameMixed')
        post = frame[[c for c in frame.columns if c%2 == 0]].sum(axis=1)
        assert post.shape == (10000,)







class MixedFrame_300_CSV(PerfTest):

    NUMBER = 2

    @classmethod
    def pd(cls) -> None:
        data_fp = SampleData.get('data_csv_fp')
        post = pd.read_csv(data_fp)
        assert post.shape == (1654482, 19)

    @classmethod
    def sf(cls) -> None:
        data_fp = SampleData.get('data_csv_fp')
        post = sf.Frame.from_csv(data_fp)
        assert post.shape == (1654482, 19)




class MixedFrame_301_JSON(PerfTest):

    NUMBER = 6

    @classmethod
    def pd(cls) -> None:
        data_url = SampleData.get('data_json_url')
        post = pd.read_json(data_url)
        assert post.shape == (5000, 5)

    @classmethod
    def sf(cls) -> None:
        data_url = SampleData.get('data_json_url')
        post = sf.Frame.from_json_url(data_url)
        assert post.shape == (5000, 5)


class MixedFrame_302_CSVHybrid(PerfTest):

    NUMBER = 2

    @classmethod
    def pd(cls) -> None:
        data_fp = SampleData.get('data_csv_fp')
        post = pd.read_csv(data_fp)
        assert post.shape == (1654482, 19)

    @classmethod
    def sf(cls) -> None:
        data_fp = SampleData.get('data_csv_fp')
        post = sf.Frame.from_pandas(pd.read_csv(data_fp))
        assert post.shape == (1654482, 19)





import typing as tp
from collections import deque
from itertools import zip_longest
from itertools import repeat
from copy import deepcopy

import numpy as np
from arraykit import resolve_dtype_iter


from static_frame.core.array_go import ArrayGO
from static_frame.core.container_util import index_from_optional_constructor
from static_frame.core.doc_str import doc_inject
from static_frame.core.exception import ErrorInitIndexLevel
from static_frame.core.hloc import HLoc
from static_frame.core.index import ILoc
from static_frame.core.index import Index
from static_frame.core.index import IndexGO
from static_frame.core.index import LocMap
from static_frame.core.index import mutable_immutable_index_filter
from static_frame.core.index_base import IndexBase
from static_frame.core.type_blocks import TypeBlocks
from static_frame.core.util import GetItemKeyType
from static_frame.core.util import GetItemKeyTypeCompound
from static_frame.core.util import IndexConstructor
from static_frame.core.util import IndexConstructors
from static_frame.core.util import IndexInitializer
from static_frame.core.util import INT_TYPES
from static_frame.core.util import DTYPE_BOOL
from static_frame.core.util import KEY_ITERABLE_TYPES
from static_frame.core.util import KEY_MULTIPLE_TYPES
from static_frame.core.util import EMPTY_TUPLE
# from static_frame.core.exception import LocInvalid


INDEX_LEVEL_SLOTS = (
            'index',
            'targets',
            'offset',
            '_depth',
            '_length',
            )

class IndexLevel:
    '''
    A nestable representation of an Index, where labels in that index optionally point to other Index objects.
    '''
    __slots__ = INDEX_LEVEL_SLOTS
    index: Index
    targets: tp.Optional[ArrayGO]
    offset: int
    _depth: tp.Optional[int]
    _length: tp.Optional[int]

    STATIC: bool = True
    _INDEX_CONSTRUCTOR = Index

    @classmethod
    def from_level_data(cls,
            level_data: tp.Any,
            get_index: tp.Callable[[IndexInitializer, int], IndexBase],
            offset: int = 0,
            depth: int = 0,
            ) -> 'IndexLevel':
        '''
        Recursive function used in ``from_tree`` constructor.
        '''
        if isinstance(level_data, dict):
            level_labels = []
            targets = np.empty(len(level_data), dtype=object)
            offset_local = 0

            # ordered key, value pairs, where the key is the label, the value is a list or dictionary; enmerate for insertion pre-allocated object array
            for idx, (k, v) in enumerate(level_data.items()):
                level_labels.append(k)
                level = cls.from_level_data(v,
                        get_index,
                        offset=offset_local,
                        depth=depth + 1)
                targets[idx] = level
                offset_local += len(level) # for lower level offsetting

            index = get_index(level_labels, depth)
            targets = ArrayGO(targets, own_iterable=True)

        else: # an iterable, terminal node, no offsets needed
            index = get_index(level_data, depth)
            targets = None

        return cls(
                index=index, #type: ignore
                offset=offset,
                targets=targets,
                )

    @classmethod
    def from_tree(cls,
            tree: tp.Any, # recursively defined
            index_constructors: tp.Optional[IndexConstructors] = None,
            depth_reference: tp.Optional[int] = None,
            ) -> 'IndexLevel':
        '''
        Convert a tree structure to an IndexLevel instance. As a tree structure is a dictionary of keys to either a sequence of hashables or a dict of other keys, there is no way to represent a zero length, non-zero depth structure.
        '''
        # tree: tp.Dict[tp.Hashable, tp.Union[Sequence[tp.Hashable], tp.Dict]]

        def get_index(labels: IndexInitializer, depth: int) -> Index:
            explicit_constructor: tp.Optional[IndexConstructor]

            if index_constructors is not None:
                explicit_constructor = index_constructors[depth]
            else:
                explicit_constructor = None

            return index_from_optional_constructor(labels, #type: ignore
                    default_constructor=cls._INDEX_CONSTRUCTOR,
                    explicit_constructor=explicit_constructor)

        if len(tree) == 0:
            return cls(get_index((), 0), depth_reference=depth_reference)
        # NOTE: code check that returned object has depth equal depth_reference
        return cls.from_level_data(tree, get_index)


    @classmethod
    def from_depth(cls, depth: int) -> 'IndexLevel':
        '''
        Create zero-legnth IndexLevel from depth.
        '''
        return cls(cls._INDEX_CONSTRUCTOR(EMPTY_TUPLE),
                own_index=True,
                depth_reference=depth,
                )

    def __init__(self,
            index: Index,
            targets: tp.Optional[ArrayGO] = None,
            offset: int = 0,
            own_index: bool = False,
            depth_reference: tp.Optional[int] = None,
            ):
        '''
        Args:
            index: a 1D Index defining outer-most labels to integers in the `targets` ArrayGO.
            offset: integer offset for this level.
            targets: None, or an ArrayGO of IndexLevel objects
            own_index: Boolean to determine whether the Index can be owned by this IndexLevel; if False, a static index will be reused if appropriate for this IndexLevel class.
            depth_reference: for zero length Levels, provide the depth if it is greater than 1.
        '''
        if not isinstance(index, Index) or index.depth > 1:
            raise ErrorInitIndexLevel('cannot create an IndexLevel from a higher-dimensional Index.')

        # NOTE: indices that contain tuples will take additional work to support; we are not at this time checking for them, though values_at_depth will fail

        if own_index:
            self.index = index
        else:
            self.index = mutable_immutable_index_filter(self.STATIC, index) #type: ignore

        self.targets = targets
        self.offset = offset

        # NOTE: once _depth is set, is ever re-evaluated over the life of the instance, even if it is an IndexLevelGO
        if len(index) > 0:
            self._depth = 1 if self.targets is None else None
            self._length = None
        else:
            if depth_reference is None:
                raise ErrorInitIndexLevel('zero length index requires specification of depth_reference')
            self._depth = depth_reference
            self._length = 0

    #---------------------------------------------------------------------------
    def __deepcopy__(self, memo: tp.Dict[int, tp.Any]) -> 'IndexLevel':
        obj = self.__new__(self.__class__)
        obj.index = deepcopy(self.index, memo)
        obj.targets = deepcopy(self.targets, memo) # ArrayGO implements __deepcopy__
        obj.offset = self.offset
        obj._depth = self._depth
        obj._length = self._length

        memo[id(self)] = obj
        return obj #type: ignore


    #---------------------------------------------------------------------------
    def depths(self) -> tp.Iterator[int]:
        '''
        Get the depth of all leaves (which should all be the same). Mostly for integrity validation.
        '''
        # NOTE: as this uses a list instead of deque, the depths given will not be in the order of the actual leaves
        if self.targets is None or not len(self.targets):
            # NOTE: requires being set in __init__
            yield self._depth #type: ignore
        else:
            levels = [(self, 0)]
            while levels:
                level, depth = levels.pop()
                if level.targets is None: # terminus
                    yield depth + 1
                else:
                    next_depth = depth + 1
                    levels.extend([(lvl, next_depth) for lvl in level.targets])

    def _get_depth(self) -> int:
        '''
        Called once over the life of an instance to set self._depth; this is not re-evaluated over the life of the instance.
        '''
        if not len(self.index):
            raise AssertionError('zero-length indices should have depth set through depth_reference')

        if self.targets is None: # this may not need to be here
            return 1

        # if we need to recurse to the max depth
        level, depth = self, 1
        while True:
            if level.targets is None: # terminus
                return depth
            level, depth = level.targets[0], depth + 1

    @property
    def depth(self) -> int:
        if self._depth is None:
            self._depth = self._get_depth()
        return self._depth

    def _get_length(self) -> int:
        # NOTE: unlike depth, length can change in an IndexLevelGO and should be re-evaluated.
        if self.targets is None or not len(self.targets):
            return self.index.__len__()

        count = 0
        levels = [self]
        while levels:
            level = levels.pop()
            if level.targets is None: # terminus
                count += level.index.__len__()
            else:
                levels.extend(level.targets)
        return count

    def __len__(self) -> int:
        '''
        The length is the sum of all leaves
        '''
        # NOTE: in IndexLevelGO, setting length to None is how length is reset after mutation.
        if self._length is None:
            self._length = self._get_length()
        return self._length

    #---------------------------------------------------------------------------
    def label_widths_at_depth(self,
            depth_level: int = 0
            ) -> tp.Iterator[tp.Tuple[tp.Hashable, int]]:
        '''
        Generator of pairs of label, width, for all labels found at a specified level.
        '''
        # given a: 1, 2, b: 1, 2, return ('a', 2), ('b', 2)

        def get_widths(index: Index,
                targets: tp.Optional[ArrayGO]
                ) -> tp.Iterator[tp.Tuple[tp.Hashable, int]]:
            if targets is None:
                for label in index:
                    yield (label, 1)
            else: # observe the offsets of the next
                transversed = 0
                for i, (label, level_next) in enumerate(
                        zip_longest(index, targets[1:], fillvalue=None)
                        ):
                    if level_next is not None:
                        # if the next offset is zero, we are moving to a component that is under a fresh hierarchy
                        if level_next.offset > 0:
                            delta = level_next.offset - transversed
                        else:
                            delta = len(targets[i])
                        yield label, delta
                        # get only the incremental addition for this label
                        transversed += delta
                    else:
                        # we cannot use offset; get length of component Levels
                        yield label, len(targets[i])

        levels = deque(((self, 0),))
        while levels:
            level, depth = levels.popleft()
            if depth == depth_level:
                yield from get_widths(level.index, level.targets)
                continue # do not need to descend
            if level.targets is not None: # terminus
                next_depth = depth + 1
                levels.extend((lvl, next_depth) for lvl in level.targets)

    def labels_at_depth(self,
            depth_level: int = 0
            ) -> tp.Iterator[tp.Hashable]:
        '''
        Generator of all labels found at a specified level, realized at full length.
        '''
        def get_labels(index: Index,
                targets: tp.Optional[ArrayGO]
                ) -> tp.Iterator[tp.Hashable]:
            if targets is None:
                yield from index
            else: # observe the offsets of the next
                transversed = 0
                for i, (label, level_next) in enumerate(
                        zip_longest(index, targets[1:], fillvalue=None)
                        ):
                    if level_next is not None:
                        if level_next.offset > 0:
                            delta = level_next.offset - transversed
                        else:
                            delta = len(targets[i])
                        yield from repeat(label, delta)
                        transversed += delta
                    else:
                        yield from repeat(label, len(targets[i]))

        levels = deque(((self, 0),))
        while levels:
            level, depth = levels.popleft()
            if depth == depth_level:
                yield from get_labels(level.index, level.targets)
                continue # do not need to descend
            if level.targets is not None: # terminus
                next_depth = depth + 1
                levels.extend((lvl, next_depth) for lvl in level.targets)


    def index_array_at_depth(self,
            depth_level: int = 0
            ) -> tp.Iterator[np.ndarray]:
        '''
        Generator of arrays found at a depth level. This does not provide full-width representation.
        '''
        levels = deque(((self, 0),))
        while levels:
            level, depth = levels.popleft()
            if depth == depth_level:
                yield level.index.values
                continue # do not need to descend
            if level.targets is not None: # terminus
                next_depth = depth + 1
                levels.extend([(lvl, next_depth) for lvl in level.targets])


    def index_at_depth(self, depth_level: int) -> tp.Iterator[tp.Hashable]:
        '''Given a depth position, iterate over label nodes at that depth. Only nodes will be provided, which for outer depths may not be of length equal to the entire index.
        '''
        if depth_level == 0:
            yield from self.index
        else:
            levels = deque(((self, 0),))
            while levels:
                level, depth = levels.popleft()
                if depth == depth_level:
                    yield from level.index
                    continue # do not need to descend
                if level.targets is not None: # terminus
                    next_depth = depth + 1
                    levels.extend([(lvl, next_depth) for lvl in level.targets])


    # TODO: consider a different name; was dtypes()
    def dtypes_iter(self) -> tp.Iterator[np.dtype]:
        '''Return an iterator of all dtypes from every depth level.'''
        if self.targets is None or not len(self.targets):
            yield self.index.values.dtype
        else:
            levels = [self]
            while levels:
                level = levels.pop()
                yield level.index.values.dtype
                if level.targets is not None: # not terminus
                    levels.extend(level.targets)


    def dtypes_at_depth(self, depth_level: int) -> tp.Iterator[np.dtype]:
        '''
        Return all dtypes found on a depth.
        '''
        if not self.index.__len__():
            if depth_level in range(self.depth):
                yield self.index.dtype
            else:
                raise RuntimeError(f'invalid depth: {depth_level}')
        else:
            levels = deque(((self, 0),))
            while levels:
                level, depth = levels.popleft()
                if depth == depth_level:
                    yield level.index.dtype
                    continue # do not need to descend
                if level.targets is not None: # terminus
                    next_depth = depth + 1
                    levels.extend([(lvl, next_depth) for lvl in level.targets])

    def dtype_per_depth(self) -> tp.Iterator[np.dtype]:
        '''Return a tuple of resolved dtypes, one from each depth level.'''
        depth_count = self.depth
        for d in range(depth_count):
            yield resolve_dtype_iter(self.dtypes_at_depth(d))

    # consider renaming index_types_per_depth
    def index_types(self) -> tp.Iterator[np.dtype]:
        '''Return an iterator of representative Index classes, one from each depth level.'''
        if not self.index.__len__():
            yield from (self.index.__class__ for _ in range(self.depth))
        elif self.targets is None:
            yield self.index.__class__
        else:
            levels = [self]
            while levels:
                level = levels.pop()
                yield level.index.__class__
                if level.targets is not None: # not terminus
                    levels.append(level.targets[0])
                else:
                    break

    #---------------------------------------------------------------------------
    def __contains__(self, key: tp.Iterable[tp.Hashable]) -> bool:
        '''Given an iterable of single-element level keys (a leaf loc), return a bool.
        '''
        if not hasattr(key, '__iter__') or isinstance(key, str):
            return False

        node = self
        for k in key:
            if not node.index.__contains__(k):
                return False

            if node.targets is not None:
                node = node.targets[node.index._loc_to_iloc(k)]
                continue

            node.index._loc_to_iloc(k)
            return True # if above does not raise

        return False

    def leaf_loc_to_iloc(self,
            key: tp.Union[tp.Iterable[tp.Hashable], ILoc, HLoc]
            ) -> int:
        '''Given an iterable of single-element level keys (a leaf loc), return the iloc value.

        Note that key components (level selectors) cannot be slices, lists, or np.ndarray.
        '''
        if isinstance(key, ILoc):
            return key.key #type: ignore [return-value]

        node = self
        pos = 0
        key_depth_max = len(key) - 1 #type: ignore

        # NOTE: rather than a for/enumerate, this could use a while loop on an iter() and explicitly look at next() results to determine if the key matches
        for key_depth, k in enumerate(key):
            if isinstance(k, KEY_MULTIPLE_TYPES):
                raise RuntimeError(f'slices cannot be used in a leaf selection into an IndexHierarchy; try HLoc[{key}].')
            if node.targets is not None:
                node = node.targets[node.index._loc_to_iloc(k)]
                pos += node.offset
            else: # targets is None, meaning we are at max depth
                # k returns an integer
                offset = node.index._loc_to_iloc(k)
                assert isinstance(offset, INT_TYPES) # enforces leaf loc
                if key_depth == key_depth_max:
                    return pos + offset
                break # return exception below if key_depth not max depth

        raise KeyError(f'Invalid key length {key_depth_max + 1}; must be length {self.depth}.')

    def loc_to_iloc(self, key: GetItemKeyTypeCompound) -> GetItemKeyType:
        '''
        This is the low-level loc_to_iloc, analagous to LocMap.loc_to_iloc as used by Index. As such, the key at this point should not be a Series or Index object.

        If key is an np.ndarray, a Boolean array will be passed through; otherwise, it will be treated as an iterable of values to be passed to leaf_loc_to_iloc.
        '''
        from static_frame.core.series import Series

        if isinstance(key, slice):
            # given a top-level definition of a slice (and if that slice results in a single value), we can get a value range
            return slice(*LocMap.map_slice_args(self.leaf_loc_to_iloc, key))

        if isinstance(key, KEY_ITERABLE_TYPES): # iterables of leaf-locs
            if key.__class__ is np.ndarray and key.dtype == bool: #type: ignore
                return key # keep as Boolean
            return [self.leaf_loc_to_iloc(x) for x in key]

        if not isinstance(key, HLoc): # assume a leaf loc tuple
            if not isinstance(key, tuple):
                raise KeyError(f'{key} cannot be used for loc selection from IndexHierarchy; try HLoc')
            return self.leaf_loc_to_iloc(key)

        # HLoc following: collect all ilocs for all leaf indices matching HLoc patterns
        ilocs = []
        levels = deque(((self, 0, 0),)) # order matters

        while levels:
            level, depth, offset = levels.popleft()

            depth_key = key[depth]
            # NOTE: depth_key should not be Series or Index at this point; IndexHierarchy is responsible for unpacking / reindexing prior to this call
            next_offset = offset + level.offset

            if depth_key.__class__ is np.ndarray and depth_key.dtype == DTYPE_BOOL: #type: ignore
                # NOTE: use length of level, not length of index, as need to observe all leafs covered at this node.
                depth_key = depth_key[next_offset: next_offset + len(level)] #type: ignore
                if len(depth_key) > len(level.index): #type: ignore
                    # given leaf-Boolean, determine what upper nodes to select
                    depth_key = level.values_at_depth(0)[depth_key]
                    if len(depth_key) > 1:
                        # NOTE: must strip repeated labels, but cannot us np.unique as must retain order
                        depth_key = list(dict.fromkeys(depth_key).keys())

            # print(level, depth, offset, depth_key, next_offset)
            if level.targets is None:
                try:
                    # NOTE: as a selection list might be given within the HLoc, it will be tested accross many indices, and should support a partial matching
                    ilocs.append(level.index._loc_to_iloc(
                            depth_key,
                            offset=next_offset,
                            partial_selection=True,
                            ))
                except KeyError:
                    pass
            else: # when not at a leaf, we are selecting level_targets to descend withing
                try: # NOTE: no offset necessary as not a leaf selection
                    iloc = level.index._loc_to_iloc(depth_key, partial_selection=True)
                except KeyError:
                    pass
                else:
                    level_targets = level.targets[iloc] # get one or more IndexLevel objects
                    next_depth = depth + 1
                    # if not an ndarray, iloc has extracted a single IndexLevel
                    if isinstance(level_targets, IndexLevel):
                        levels.append((level_targets, next_depth, next_offset))
                    else:
                        levels.extend((lvl, next_depth, next_offset)
                                for lvl in level_targets)

        iloc_count = len(ilocs)
        if iloc_count == 0:
            raise KeyError('no matching keys across all levels')

        if iloc_count == 1 and not key.has_key_multiple():
            return ilocs[0] # drop to a single iloc selection

        # NOTE: might be able to combine contiguous ilocs into a single slice
        iloc_flat: tp.List[GetItemKeyType] = [] # combine into one flat iloc
        length = self.__len__()
        for part in ilocs:
            if isinstance(part, slice):
                iloc_flat.extend(range(*part.indices(length)))
            elif isinstance(part, INT_TYPES):
                iloc_flat.append(part)
            else: # assume it is an iterable
                iloc_flat.extend(part) #type: ignore
        return iloc_flat

    #---------------------------------------------------------------------------
    @property
    def values(self) -> np.ndarray:
        '''
        Return a single immutable NumPy 2D array of all labels found in this IndexLevels instance. This may coerce types.
        '''
        depth_count = self.depth
        shape = self.__len__(), depth_count

        # need to get a compatible dtype for all dtypes
        dtype = resolve_dtype_iter(self.dtypes_iter())
        labels = np.empty(shape, dtype=dtype)

        row_count = 0
        levels: tp.Deque[tp.Tuple['IndexLevel', int, tp.Optional[np.ndarray]]]
        levels = deque(((self, 0, None),)) # order matters

        while levels:
            level, depth, row_previous = levels.popleft()

            if level.targets is None:
                rows = len(level.index.values)
                row_slice = slice(row_count, row_count + rows)
                labels[row_slice, :] = row_previous
                labels[row_slice, depth] = level.index.values
                row_count += rows

            else: # target is iterable np.ndaarray
                depth_next = depth + 1
                for label, level_target in zip(level.index.values, level.targets):
                    if row_previous is None:
                        # shown to be faster to allocate entire row width
                        row = np.empty(depth_count, dtype=dtype)
                    else:
                        row = row_previous.copy()
                    row[depth] = label
                    levels.append((level_target, depth_next, row))

        labels.flags.writeable = False
        return labels

    def __iter__(self) -> tp.Iterator[tp.Tuple[tp.Hashable, ...]]:
        # NOTE: this implementation shown to be faster than a purely recursive implementation.
        depth_count = self.depth
        levels: tp.Deque[tp.Tuple['IndexLevel', int, tp.Optional[np.ndarray]]]
        levels = deque(((self, 0, None),)) # order matters

        while levels:
            level, depth, row_previous = levels.popleft()

            if level.targets is None:
                for v in level.index.values:
                    row_previous[depth] = v #type: ignore
                    yield tuple(row_previous) #type: ignore
            else: # target is iterable np.ndaarray
                depth_next = depth + 1
                for label, level_target in zip(level.index.values, level.targets):
                    if row_previous is None:
                        # shown to be faster to allocate entire row width
                        row = [None] * depth_count
                    else:
                        row = row_previous.copy()
                    row[depth] = label
                    levels.append((level_target, depth_next, row))


    def values_at_depth(self,
            depth_level: int
            ) -> np.ndarray:
        '''
        For the given depth, return a correctly typed immutable array of length equal to the number of rows in the cosolidate values presentation.
        '''
        depth_count = self.depth
        dtype = tuple(self.dtype_per_depth())[depth_level]

        length = self.__len__()
        # pre allocate array to ensure we use a resolved type
        array = np.empty(length, dtype=dtype)

        if length == 0:
            array.flags.writeable = False
            return array

        if depth_level == depth_count - 1:
            # at maximal depth, can concat underlying arrays
            np.concatenate(
                    tuple(self.index_array_at_depth(depth_level)),
                    out=array
                    )
        else:
            arrays = []
            for value, size in self.label_widths_at_depth(
                    depth_level=depth_level):
                if dtype.kind == 'O' and isinstance(value, tuple):
                    # this appears to the only way to do this:
                    part = np.empty(size, dtype=dtype)
                    for i in range(size):
                        part[i] = value
                    arrays.append(part)
                else:
                    arrays.append(np.full(size, value, dtype=dtype))

            np.concatenate(arrays, out=array)

            #NOTE: This alternative form produced a unicode error only on Windows up ot NP 1.17 for some tests
            # start = 0
            # for value, size in self.label_widths_at_depth(depth_level):
            #     end = start + size
            #     array[start: end] = value
            #     start = end

        array.flags.writeable = False
        return array

    #---------------------------------------------------------------------------
    @doc_inject()
    def equals(self,
            other: tp.Any,
            *,
            compare_name: bool = False,
            compare_dtype: bool = False,
            compare_class: bool = False,
            skipna: bool = True,
            ) -> bool:
        '''
        {doc}

        Args:
            {compare_name}
            {compare_dtype}
            {compare_class}
            {skipna}
        '''
        if id(other) == id(self):
            return True

        if compare_class and self.__class__ != other.__class__:
            return False
        elif not isinstance(other, IndexLevel):
            return False

        # same type from here
        if self.__len__() != other.__len__():
            return False
        if self.depth != other.depth:
            return False

        kwargs = dict(
                compare_name=compare_name,
                compare_dtype=compare_dtype,
                compare_class=compare_class,
                skipna=skipna,
                )

        if ((self.targets is None or len(self.targets) == 0)
                and (other.targets is None or len(other.targets) == 0)):
            return self.index.equals(other.index, **kwargs) #type: ignore

        # same length and depth, can traverse trees
        # can store tuple of object ids to note those that have already been examined.
        equal_pairs = set()

        levels_self = [self]
        levels_other = [other]
        while levels_self and levels_other:
            level_self = levels_self.pop()
            level_other = levels_other.pop()

            pair = (id(level_self.index), id(level_other.index))
            pair_found = pair in equal_pairs

            if not pair_found and not level_self.index.equals(level_other.index, **kwargs):
                return False

            if not pair_found: # but we know it is equal
                equal_pairs.add(pair)

            if level_self.targets is not None and level_other.targets is not None: # not terminus
                levels_self.extend(level_self.targets)
                levels_other.extend(level_other.targets)
            if level_self.targets is None and level_other.targets is None: # terminus
                continue
            if level_self.targets is None or level_other.targets is None:
                # at least one is at a terminus
                return False

        if not levels_self and not levels_other:
            return True # both exhausted
        return False #pragma: no cover

    #---------------------------------------------------------------------------
    # exporters

    def to_index_level(self,
            offset: tp.Optional[int] = 0,
            cls: tp.Optional[tp.Type['IndexLevel']] = None,
            ) -> 'IndexLevel':
        '''
        A deepcopy with optional adjustments, such as a different offset and possibly a different class. The supplied class will be used to construct the IndexLevel instance (as well as internal indices), permitting the production of an IndexLevelGO.

        Args:
            offset: optionally provide a new offset for the copy. This is not applied recursively
        '''
        cls = cls if cls else self.__class__

        index = mutable_immutable_index_filter(cls.STATIC, self.index)

        if self.targets is not None:
            targets: tp.Optional[ArrayGO] = ArrayGO(
                [t.to_index_level(offset=None, cls=cls) for t in self.targets],
                own_iterable=True)
        else:
            targets = None

        offset = self.offset if offset is None else offset
        return cls(index=index, #type: ignore
                targets=targets,
                offset=offset,
                depth_reference=self.depth,
                )

    def to_type_blocks(self) -> TypeBlocks:
        '''
        Provide a correctly typed TypeBlocks representation.
        '''
        try:
            depth_count = self.depth
        except StopIteration:
            # assume we have no depth or length
            return TypeBlocks.from_zero_size_shape()

        return TypeBlocks.from_blocks(
                self.values_at_depth(d) for d in range(depth_count)
                )

#-------------------------------------------------------------------------------
class IndexLevelGO(IndexLevel):
    '''Grow only variant of IndexLevel
    '''
    __slots__ = INDEX_LEVEL_SLOTS
    index: IndexGO
    targets: tp.Optional[np.ndarray]
    offset: int
    _depth: tp.Optional[int]
    _length: tp.Optional[int]

    STATIC: bool = False
    _INDEX_CONSTRUCTOR = IndexGO

    #---------------------------------------------------------------------------
    # grow only mutation
    # depth cannot change over the life of IndexLevel

    def extend(self, level: IndexLevel) -> None:
        '''Extend this IndexLevel with another IndexLevel, assuming that it has compatible depth and Index types.
        '''
        depth = self.depth

        if level.targets is None:
            raise RuntimeError('found IndexLevel with None as targets')
        if depth != level.depth:
            raise RuntimeError('level for extension does not have necessary levels.')

        for type_self, type_other in zip(self.index_types(), level.index_types()):
            # self type should be a GO; if other is non-GO, or GO, we can accept it
            if not issubclass(type_self, type_other):
                raise RuntimeError(f'level for extension does not have corresponding types: {type_self}, {type_other}')

        # this will raise for duplicates
        self.index.extend(level.index.values)

        def target_gen() -> tp.Iterator[GetItemKeyType]:
            offset_prior = self.__len__()
            for t in level.targets: #type: ignore
                # only need to update offsets at this level, as lower levels are relative to this
                target = t.to_index_level(offset_prior, cls=self.__class__)
                offset_prior += len(target)
                yield target

        if self.targets is None:
            raise RuntimeError('found IndexLevel with None as targets')

        self.targets.extend(target_gen())

        # defer calculation be setting _length to None
        self._length = None

    def append(self, key: tp.Sequence[tp.Hashable]) -> None:
        '''Add a single, full-depth leaf loc.
        '''
        # find fist depth that does not contain key
        depth_count = self.depth

        if len(key) != depth_count:
            raise RuntimeError('appending key {} of insufficent depth {}'.format(
                        key, depth_count))

        level_previous: tp.Optional['IndexLevelGO']

        if not self.index.__len__():
            # where we have zero length, create new root index and targets from the key alone
            depth_max = depth_count - 1
            level_previous = None

            for depth in range(depth_max, -1, -1):
                k = key[depth]
                # NOTE: we do not want to take index_types, as it based on notional index from zero length structure
                # index_constructor = index_types[depth]
                index = self._INDEX_CONSTRUCTOR((k,))
                if depth == depth_max:
                    targets = None
                else:
                    targets = ArrayGO([level_previous,], own_iterable=True)

                if depth != 0:
                    level_previous = IndexLevelGO(index, targets)
                else:
                    self.index = index
                    self.targets = targets
            self._length = None
            return

        # NOTE: does not use index_types when starting from a zero-length IndexLevels
        index_types = tuple(self.index_types())

        depth_not_found = -1
        edge_nodes = np.empty(depth_count, dtype=object)
        node = self

        for depth, k in enumerate(key):
            edge_nodes[depth] = node
            # only set on first encounter in descent
            if depth_not_found == -1 and not node.index.__contains__(k):
                depth_not_found = depth
            if node.targets is not None:
                node = node.targets[-1]

        if depth_not_found == -1:
            raise RuntimeError('unable to set depth_not_found') #pragma: no cover

        level_previous = None

        # iterate from the innermost depth out
        for depth in range(depth_count - 1, depth_not_found - 1, -1):
            node = edge_nodes[depth]
            k = key[depth]

            if depth == depth_not_found:
                # when at the the depth not found, we always update the index
                node.index.append(k)
                # if we have targets, must update them
                if node.targets is not None:
                    assert level_previous is not None
                    level_previous.offset = node.__len__() #type: ignore [unreachable]
                    node.targets.append(level_previous)
            else: # depth not found is higher up
                # NOTE: do not need to use index_from_optional_constructor, as no explicit constructor is being supplied, and we can expect that the existing types must be valid
                index_constructor = index_types[depth]
                if node.targets is None:
                    # we are at the max depth; will need to create a LevelGO to append in the next level
                    level_previous = IndexLevelGO(
                            index=index_constructor((k,)),
                            offset=0,
                            targets=None
                            )
                else:
                    targets = ArrayGO([level_previous,], own_iterable=True)
                    level_previous = IndexLevelGO(
                            index=index_constructor((k,)),
                            offset=0,
                            targets=targets
                            )
        # defer calculation be setting _length to None for all edge levels
        for node in edge_nodes:
            node._length = None




